from typing import Union, List, Tuple

import pytest
import torch
from torch_geometric.data import Data
from torch_geometric.utils import to_dense_adj, coalesce

from pydgn.data.dataset import DatasetInterface, OGBGDatasetInterface
from pydgn.data.splitter import (
    Splitter,
    SingleGraphSplitter,
    LinkPredictionSingleGraphSplitter,
)


class FakeGraphClassificationDataset(DatasetInterface):
    def __init__(self, root=None, name="mock_dataset"):
        super().__init__(root, name)
        self.num_samples = 1000
        self.feats = 10
        self.classes = 3
        self.data = []

        for s in range(self.num_samples):
            num_nodes = int(torch.randint(low=2, high=100, size=(1,)))
            x = torch.rand((num_nodes, self.feats))
            y = torch.randint(self.classes, (1,))
            edge_index = torch.randint(num_nodes, (2, num_nodes * 2))
            edge_attr = torch.rand(edge_index.shape[1])

            self.data.append(
                Data(x=x, edge_index=edge_index, edge_attr=edge_attr, y=y)
            )

    def get(self, idx: int) -> Data:
        return self.data[idx]

    @property
    def dim_node_features(self) -> int:
        return self.data[0].x.shape[1]

    @property
    def dim_edge_features(self) -> int:
        return 1

    @property
    def dim_target(self) -> int:
        return self.classes

    def len(self) -> int:
        return len(self)

    def __len__(self) -> int:
        return len(self.data)


@pytest.fixture
def graph_classification_dataset():
    """
    Builds a random dataset for graph classification
    """
    return FakeGraphClassificationDataset()


@pytest.fixture
def node_classification_dataset():
    """
    Builds a random dataset for node classification (single graph)
    """

    class FakeDataset(DatasetInterface):
        def __init__(self):
            super().__init__(None, None)
            self.feats = 10
            self.classes = 3

            num_nodes = int(torch.randint(low=1000, high=2000, size=(1,)))
            x = torch.rand((num_nodes, self.feats))
            y = torch.randint(self.classes, (num_nodes,))
            edge_index = torch.randint(num_nodes, (2, num_nodes * 2))
            edge_attr = torch.rand(edge_index.shape[1])

            self.data = Data(
                x=x, edge_index=edge_index, edge_attr=edge_attr, y=y
            )

        def get(self, idx: int) -> Data:
            return self.data

        @property
        def dim_node_features(self) -> int:
            return self.data[0].x.shape[1]

        @property
        def dim_edge_features(self) -> int:
            return 1

        @property
        def dim_target(self) -> int:
            return self.classes

        def len(self) -> int:
            return len(self)

        def __len__(self) -> int:
            return 1

    return FakeDataset()


@pytest.fixture
def link_prediction_dataset():
    """
    Builds a random dataset for link prediction (single graph)
    """

    class FakeDataset(DatasetInterface):
        def __init__(self):
            super().__init__(None, None)
            self.feats = 10
            self.classes = 3

            num_nodes = int(torch.randint(low=100, high=200, size=(1,)))
            x = torch.rand((num_nodes, self.feats))
            # add one class per link, just in case (future link class. tasks)
            y = torch.randint(self.classes, (num_nodes * 2,))
            edge_index = torch.randint(num_nodes, (2, num_nodes * 2))
            edge_attr = torch.rand(edge_index.shape[1])

            # Very important line of code: remove duplicates
            edge_index, edge_attr = coalesce(
                edge_index, edge_attr, num_nodes=num_nodes
            )

            self.data = Data(
                x=x, edge_index=edge_index, edge_attr=edge_attr, y=y
            )

        def get(self, idx: int) -> Data:
            return self.data

        @property
        def dim_node_features(self) -> int:
            return self.data[0].x.shape[1]

        @property
        def dim_edge_features(self) -> int:
            return 1

        @property
        def dim_target(self) -> int:
            return self.classes

        def len(self) -> int:
            return len(self)

        def __len__(self) -> int:
            return 1

    return FakeDataset()


# To each task its own splitter
@pytest.fixture
def node_and_graph_task_input(
    node_classification_dataset, graph_classification_dataset
):
    """
    Returns tuples (dataset, splitter) to test for data splits overlap
    """
    return [
        (graph_classification_dataset, Splitter),
        (node_classification_dataset, SingleGraphSplitter),
    ]


def test_node_graph_split_overlap(node_and_graph_task_input):
    """
    Tests data splits overlap for node and graph prediction
    """
    for dataset, splitter_class in node_and_graph_task_input:
        for n_outer_folds in [1, 10]:
            for n_inner_folds in [1, 10]:
                for stratify in [False, True]:

                    splitter = splitter_class(
                        n_outer_folds=n_outer_folds,
                        n_inner_folds=n_inner_folds,
                        seed=0,
                        stratify=stratify,
                        shuffle=True,
                        outer_val_ratio=0.1,
                    )

                    _, targets = splitter.get_targets(dataset)
                    splitter.split(dataset, targets)

                    for outer in range(n_outer_folds):

                        outer_train_idxs = splitter.outer_folds[
                            outer
                        ].train_idxs
                        outer_val_idxs = splitter.outer_folds[outer].val_idxs
                        outer_test_idxs = splitter.outer_folds[outer].test_idxs

                        # False if empty
                        assert not bool(
                            set(outer_train_idxs)
                            & set(outer_val_idxs)
                            & set(outer_test_idxs)
                        )

                        for inner in range(n_inner_folds):

                            inner_train_idxs = splitter.inner_folds[outer][
                                inner
                            ].train_idxs
                            inner_val_idxs = splitter.inner_folds[outer][
                                inner
                            ].val_idxs

                            # False if empty
                            assert not bool(
                                set(inner_train_idxs)
                                & set(inner_val_idxs)
                                & set(outer_test_idxs)
                            )

                            # Check length consistency
                            if len(dataset) == 1:
                                assert (
                                    len(inner_train_idxs)
                                    + len(inner_val_idxs)
                                    + len(outer_test_idxs)
                                    == dataset[0].x.shape[0]
                                )
                            else:
                                assert len(inner_train_idxs) + len(
                                    inner_val_idxs
                                ) + len(outer_test_idxs) == len(dataset)

                        # Check length consistency
                        if len(dataset) == 1:
                            assert (
                                len(outer_train_idxs)
                                + len(outer_val_idxs)
                                + len(outer_test_idxs)
                                == dataset[0].x.shape[0]
                            )
                        else:
                            assert len(outer_train_idxs) + len(
                                outer_val_idxs
                            ) + len(outer_test_idxs) == len(dataset)


# To each task its own splitter
@pytest.fixture
def link_task_input(link_prediction_dataset):
    """
    Returns tuples (dataset, splitter) to test for data splits overlap
    """
    return [
        (link_prediction_dataset, LinkPredictionSingleGraphSplitter),
    ]


def test_link_split_overlap(link_task_input):
    """
    Tests data splits overlap for link prediction
    """

    def _run_checks(
        original_edge_index,
        pos_train_edges,
        neg_train_edges,
        pos_val_edges,
        neg_val_edges,
        pos_test_edges,
        neg_test_edges,
    ):
        """
        Runs additional memory-intensive checks to ensure the edge split is ok
        and does not contain overlaps.
        """
        original_Adj = to_dense_adj(original_edge_index)[0]

        pos_train_edges = torch.tensor(pos_train_edges).long()
        neg_train_edges = torch.tensor(neg_train_edges).long()
        pos_val_edges = torch.tensor(pos_val_edges).long()
        neg_val_edges = torch.tensor(neg_val_edges).long()
        pos_test_edges = torch.tensor(pos_test_edges).long()
        neg_test_edges = torch.tensor(neg_test_edges).long()

        # First: check positive edges perfectly reconstruct adjacency matrix
        N = original_Adj.shape[0]
        # TEST: restore A with partitioned positive links
        A_pos = to_dense_adj(pos_train_edges, max_num_nodes=N)[0, :, :]
        A_pos += to_dense_adj(pos_val_edges, max_num_nodes=N)[0, :, :]
        A_pos += to_dense_adj(pos_test_edges, max_num_nodes=N)[0, :, :]

        # It is also a way to check that positive edges do not overlap
        # If I do not call to_undirected to val and test edges when
        # undirected==True, this won't hold
        # assert torch.all(A == A_pos)

        # Second: check edges do not overlap
        # Can be done by checking the sum of positive and negative adj matrices
        # holds values smaller or equal than 1.

        # TEST: check negative links are separate from positive links
        A_neg = to_dense_adj(neg_train_edges, max_num_nodes=N)[0, :, :]
        A_neg += to_dense_adj(neg_val_edges, max_num_nodes=N)[0, :, :]
        A_neg += to_dense_adj(neg_test_edges, max_num_nodes=N)[0, :, :]

        assert torch.all((A_neg + A_pos) <= 1.0)

    for dataset, splitter_class in link_task_input:
        for n_outer_folds in [1, 10]:
            for n_inner_folds in [1, 10]:
                for stratify in [False, True]:

                    splitter = splitter_class(
                        n_outer_folds=n_outer_folds,
                        n_inner_folds=n_inner_folds,
                        seed=0,
                        stratify=stratify,
                        shuffle=True,
                        outer_val_ratio=0.1,
                        undirected=False,
                        avoid_opposite_negative_edges=True,
                    )

                    splitter.split(dataset, None)

                    for outer in range(n_outer_folds):

                        (
                            outer_pos_train_edges,
                            outer_train_attr,
                            outer_neg_train_edges,
                        ) = splitter.outer_folds[outer].train_idxs

                        (
                            outer_pos_val_edges,
                            outer_val_attr,
                            outer_neg_val_edges,
                        ) = splitter.outer_folds[outer].val_idxs

                        (
                            outer_pos_test_edges,
                            outer_test_attr,
                            outer_neg_test_edges,
                        ) = splitter.outer_folds[outer].test_idxs

                        _run_checks(
                            dataset[0].edge_index,
                            outer_pos_train_edges,
                            outer_neg_train_edges,
                            outer_pos_val_edges,
                            outer_neg_val_edges,
                            outer_pos_test_edges,
                            outer_neg_test_edges,
                        )

                        for inner in range(n_inner_folds):
                            (
                                inner_pos_train_edges,
                                inner_train_attr,
                                inner_neg_train_edges,
                            ) = splitter.inner_folds[outer][inner].train_idxs

                            (
                                inner_pos_val_edges,
                                inner_val_attr,
                                inner_neg_val_edges,
                            ) = splitter.inner_folds[outer][inner].val_idxs

                            _run_checks(
                                dataset[0].edge_index,
                                inner_pos_train_edges,
                                inner_neg_train_edges,
                                inner_pos_val_edges,
                                inner_neg_val_edges,
                                outer_pos_test_edges,
                                outer_neg_test_edges,
                            )



from torch.nn.utils import clip_grad_value_

from pydgn.experiment.util import s2c
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class GradientClipper(EventHandler):
    r"""
    GradientClipper is the main event handler for gradient clippers.
    Just pass a PyTorch scheduler together with its
    arguments in the configuration file.

    Args:
        clip_value (float): the gradient will be clipped in
            [-clip_value, clip_value]
        kwargs (dict): additional arguments
    """

    def __init__(self, clip_value: float, **kwargs: dict):
        self.clip_value = clip_value

    def on_backward(self, state: State):
        """
        Clips the gradients of the model before the weights are updated.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        clip_grad_value_(state.model.parameters(), clip_value=self.clip_value)



from copy import deepcopy
from shutil import rmtree
from unittest.mock import patch

import pytest
import torch
from torch_geometric.loader import DataLoader
from torch_geometric.utils import sort_edge_index

from pydgn.data.dataset import DatasetInterface, ToyIterableDataset
from pydgn.data.provider import (
    DataProvider,
    SingleGraphDataProvider,
    LinkPredictionSingleGraphDataProvider,
    IterableDataProvider,
)
from pydgn.data.splitter import (
    Splitter,
    SingleGraphSplitter,
    LinkPredictionSingleGraphSplitter,
)
from tests.data.test_data_splitter import (
    link_prediction_dataset,
    graph_classification_dataset,
    node_classification_dataset,
)


def mock_get_dataset(cls, **kwargs):
    """
    Returns the dataset stored in the object (see main test)
    """
    return deepcopy(cls.dataset)


def mock_get_splitter(cls, **kwargs):
    """
    Instantiates a splitter and generates random splits
    """
    if cls.splitter is None:
        splitter = Splitter(
            n_outer_folds=cls.outer_folds,
            n_inner_folds=cls.inner_folds,
            seed=0,
            stratify=True,
            shuffle=True,
            outer_val_ratio=0.1,
        )
        dataset = cls._get_dataset()
        splitter.split(dataset, splitter.get_targets(dataset)[1])
        cls.splitter = splitter
        return cls.splitter

    return cls.splitter


def mock_get_link_pred_splitter(cls, **kwargs):
    """
    Instantiates a splitter and generates random splits
    """
    if cls.splitter is None:

        splitter = LinkPredictionSingleGraphSplitter(
            n_outer_folds=cls.outer_folds,
            n_inner_folds=cls.inner_folds,
            seed=0,
            stratify=True,
            shuffle=True,
            outer_val_ratio=0.1,
        )
        dataset = cls._get_dataset()
        splitter.split(dataset, None)
        cls.splitter = splitter
        return cls.splitter

    return cls.splitter


def mock_get_singlegraphsplitter(cls, **kwargs):
    """
    Instantiates a splitter and generates random splits
    """
    splitter = SingleGraphSplitter(
        n_outer_folds=cls.outer_folds,
        n_inner_folds=cls.inner_folds,
        seed=0,
        stratify=True,
        shuffle=True,
        outer_val_ratio=0.1,
    )
    dataset = cls._get_dataset()
    splitter.split(dataset, splitter.get_targets(dataset)[1])

    return splitter


@patch.object(DataProvider, "_get_splitter", mock_get_splitter)
@patch.object(DataProvider, "_get_dataset", mock_get_dataset)
def test_DataProvider(graph_classification_dataset):
    """
    Check that the data provider returns the correct data associated
    with different data splits
    """
    batch_size = 32
    for outer_folds in [1, 10]:
        for inner_folds in [1, 10]:
            for shuffle in [False, True]:
                provider = DataProvider(
                    None,
                    None,
                    DatasetInterface,
                    "",
                    DataLoader,
                    {},
                    outer_folds=outer_folds,
                    inner_folds=inner_folds,
                )
                provider.dataset = graph_classification_dataset
                provider.set_exp_seed(0)

                for o in range(outer_folds):
                    provider.set_outer_k(o)
                    for i in range(inner_folds):
                        provider.set_inner_k(i)

                        inner_train_loader = provider.get_inner_train(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert set(inner_train_loader.dataset.indices) == set(
                            provider._get_splitter()
                            .inner_folds[o][i]
                            .train_idxs
                        )

                        inner_val_loader = provider.get_inner_val(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert set(inner_val_loader.dataset.indices) == set(
                            provider._get_splitter().inner_folds[o][i].val_idxs
                        )

                    provider.set_inner_k(None)

                    outer_train_loader = provider.get_outer_train(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(outer_train_loader.dataset.indices) == set(
                        provider._get_splitter().outer_folds[o].train_idxs
                    )

                    outer_val_loader = provider.get_outer_val(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(outer_val_loader.dataset.indices) == set(
                        provider._get_splitter().outer_folds[o].val_idxs
                    )

                    outer_test_loader = provider.get_outer_test(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(outer_test_loader.dataset.indices) == set(
                        provider._get_splitter().outer_folds[o].test_idxs
                    )


@patch.object(
    SingleGraphDataProvider, "_get_splitter", mock_get_singlegraphsplitter
)
@patch.object(SingleGraphDataProvider, "_get_dataset", mock_get_dataset)
def test_SingleGraphDataProvider(node_classification_dataset):
    """
    Check that the data provider returns the correct data associated
    with different data splits
    """
    batch_size = 32
    for outer_folds in [1, 10]:
        for inner_folds in [1, 10]:
            for shuffle in [False, True]:
                provider = SingleGraphDataProvider(
                    None,
                    None,
                    DatasetInterface,
                    "",
                    DataLoader,
                    {},
                    outer_folds=outer_folds,
                    inner_folds=inner_folds,
                )
                provider.dataset = node_classification_dataset
                provider.set_exp_seed(0)

                for o in range(outer_folds):
                    provider.set_outer_k(o)
                    for i in range(inner_folds):
                        provider.set_inner_k(i)

                        inner_train_loader = provider.get_inner_train(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert set(
                            inner_train_loader.dataset.data.training_indices.tolist()
                        ) == set(
                            provider._get_splitter()
                            .inner_folds[o][i]
                            .train_idxs
                        )

                        assert set(
                            inner_train_loader.dataset.data.eval_indices.tolist()
                        ) == set(
                            provider._get_splitter()
                            .inner_folds[o][i]
                            .train_idxs
                        )

                        inner_val_loader = provider.get_inner_val(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert set(
                            inner_val_loader.dataset.data.training_indices.tolist()
                        ) == set(
                            provider._get_splitter()
                            .inner_folds[o][i]
                            .train_idxs
                        )

                        assert set(
                            inner_val_loader.dataset.data.eval_indices.tolist()
                        ) == set(
                            provider._get_splitter().inner_folds[o][i].val_idxs
                        )

                    provider.set_inner_k(None)

                    outer_train_loader = provider.get_outer_train(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(
                        outer_train_loader.dataset.data.training_indices.tolist()
                    ) == set(
                        provider._get_splitter().outer_folds[o].train_idxs
                    )

                    assert set(
                        outer_train_loader.dataset.data.eval_indices.tolist()
                    ) == set(
                        provider._get_splitter().outer_folds[o].train_idxs
                    )

                    outer_val_loader = provider.get_outer_val(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(
                        outer_val_loader.dataset.data.training_indices.tolist()
                    ) == set(
                        provider._get_splitter().outer_folds[o].train_idxs
                    )

                    assert set(
                        outer_val_loader.dataset.data.eval_indices.tolist()
                    ) == set(provider._get_splitter().outer_folds[o].val_idxs)

                    outer_test_loader = provider.get_outer_test(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert set(
                        outer_test_loader.dataset.data.training_indices.tolist()
                    ) == set(
                        provider._get_splitter().outer_folds[o].train_idxs
                    )

                    assert set(
                        outer_test_loader.dataset.data.eval_indices.tolist()
                    ) == set(provider._get_splitter().outer_folds[o].test_idxs)


@patch.object(
    LinkPredictionSingleGraphDataProvider,
    "_get_splitter",
    mock_get_link_pred_splitter,
)
@patch.object(
    LinkPredictionSingleGraphDataProvider, "_get_dataset", mock_get_dataset
)
def test_LinkPredictionSingleGraphDataProvider(link_prediction_dataset):
    """
    Check that the data provider returns the correct data associated
    with different data splits
    """
    for batch_size in [0, 32]:
        for outer_folds in [1, 10]:
            for inner_folds in [1, 10]:
                for shuffle in [False, True]:
                    provider = LinkPredictionSingleGraphDataProvider(
                        None,
                        None,
                        DatasetInterface,
                        "",
                        DataLoader,
                        {},
                        outer_folds=outer_folds,
                        inner_folds=inner_folds,
                    )
                    provider.set_exp_seed(0)

                    for o in range(outer_folds):
                        provider.set_outer_k(o)
                        for i in range(inner_folds):

                            # The graph gets modified at each call
                            provider.dataset = link_prediction_dataset
                            provider.set_inner_k(i)

                            inner_train_loader = provider.get_inner_train(
                                shuffle=shuffle, batch_size=batch_size
                            )

                            inner_pos_train_edges, _, inner_neg_train_edges = (
                                provider._get_splitter()
                                .inner_folds[o][i]
                                .train_idxs
                            )

                            inner_val_loader = provider.get_inner_val(
                                shuffle=shuffle, batch_size=batch_size
                            )

                            inner_pos_val_edges, _, inner_neg_val_edges = (
                                provider._get_splitter()
                                .inner_folds[o][i]
                                .val_idxs
                            )

                            for loader, pos_list, neg_list in [
                                (
                                    inner_train_loader,
                                    inner_pos_train_edges,
                                    inner_neg_train_edges,
                                ),
                                (
                                    inner_val_loader,
                                    inner_pos_val_edges,
                                    inner_neg_val_edges,
                                ),
                            ]:
                                pos_edge_idx = torch.cat(
                                    [d.y[1] for d in loader.dataset], dim=1
                                )
                                neg_edge_idx = torch.cat(
                                    [d.y[2] for d in loader.dataset], dim=1
                                )

                                pos_list = torch.tensor(pos_list).long()

                                neg_list = torch.tensor(neg_list).long()

                                pos_edge_idx = sort_edge_index(pos_edge_idx)
                                pos_list = sort_edge_index(pos_list)

                                neg_edge_idx = sort_edge_index(neg_edge_idx)
                                neg_list = sort_edge_index(neg_list)

                                assert (
                                    pos_edge_idx.tolist() == pos_list.tolist()
                                )

                                assert (
                                    neg_edge_idx.tolist() == neg_list.tolist()
                                )

                        provider.set_inner_k(None)

                        outer_train_loader = provider.get_outer_train(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        outer_pos_train_edges, _, outer_neg_train_edges = (
                            provider._get_splitter().outer_folds[o].train_idxs
                        )

                        outer_val_loader = provider.get_outer_val(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        outer_pos_val_edges, _, outer_neg_val_edges = (
                            provider._get_splitter().outer_folds[o].val_idxs
                        )

                        outer_test_loader = provider.get_outer_test(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        outer_pos_test_edges, _, outer_neg_test_edges = (
                            provider._get_splitter().outer_folds[o].test_idxs
                        )

                        for loader, pos_list, neg_list in [
                            (
                                outer_train_loader,
                                outer_pos_train_edges,
                                outer_neg_train_edges,
                            ),
                            (
                                outer_val_loader,
                                outer_pos_val_edges,
                                outer_neg_val_edges,
                            ),
                            (
                                outer_test_loader,
                                outer_pos_test_edges,
                                outer_neg_test_edges,
                            ),
                        ]:
                            pos_edge_idx = torch.cat(
                                [d.y[1] for d in loader.dataset], dim=1
                            )
                            neg_edge_idx = torch.cat(
                                [d.y[2] for d in loader.dataset], dim=1
                            )

                            pos_list = torch.tensor(pos_list).long()

                            neg_list = torch.tensor(neg_list).long()

                            pos_edge_idx = sort_edge_index(pos_edge_idx)
                            pos_list = sort_edge_index(pos_list)

                            neg_edge_idx = sort_edge_index(neg_edge_idx)
                            neg_list = sort_edge_index(neg_list)

                            assert pos_edge_idx.tolist() == pos_list.tolist()

                            assert neg_edge_idx.tolist() == neg_list.tolist()


def mock_get_iterable_splitter(cls, **kwargs):
    """
    Instantiates a splitter and generates random splits
    """
    if cls.splitter is None:
        splitter = Splitter(
            n_outer_folds=cls.outer_folds,
            n_inner_folds=cls.inner_folds,
            seed=0,
            stratify=False,
            shuffle=True,
            outer_val_ratio=0.1,
        )
        dataset = cls._get_dataset()
        splitter.split(dataset, None)
        cls.splitter = splitter
        return cls.splitter

    return cls.splitter


def mock_get_iterable_dataset(cls, **kwargs):
    """
    Returns the dataset stored in the object (see main test)
    """
    data_root = "tests/data/debug/DATA"

    if "url_indices" in kwargs:
        indices = kwargs["url_indices"]
        return ToyIterableDataset(data_root, "toy", url_indices=indices)
    return ToyIterableDataset(data_root, "toy")


@patch.object(IterableDataProvider, "_get_dataset", mock_get_iterable_dataset)
@patch.object(
    IterableDataProvider, "_get_splitter", mock_get_iterable_splitter
)
def test_IterableDataProvider():
    """
    Check that the data provider returns the correct data associated
    with different data splits
    """
    data_root = "tests/data/debug/DATA"
    batch_size = 32
    for outer_folds in [1, 3]:
        for inner_folds in [1, 3]:
            for shuffle in [False, True]:
                provider = IterableDataProvider(
                    data_root,
                    None,
                    DatasetInterface,
                    "",
                    DataLoader,
                    {},
                    outer_folds=outer_folds,
                    inner_folds=inner_folds,
                )
                provider.dataset = ToyIterableDataset(data_root, "toy")
                provider.set_exp_seed(0)

                for o in range(outer_folds):
                    provider.set_outer_k(o)
                    for i in range(inner_folds):
                        provider.set_inner_k(i)

                        inner_train_loader = provider.get_inner_train(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert len(inner_train_loader.dataset) == len(
                            set(
                                provider._get_splitter()
                                .inner_folds[o][i]
                                .train_idxs
                            )
                        )

                        inner_val_loader = provider.get_inner_val(
                            shuffle=shuffle, batch_size=batch_size
                        )

                        assert len(inner_val_loader.dataset) == len(
                            set(
                                provider._get_splitter()
                                .inner_folds[o][i]
                                .val_idxs
                            )
                        )

                    provider.set_inner_k(None)

                    outer_train_loader = provider.get_outer_train(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert len(outer_train_loader.dataset) == len(
                        set(provider._get_splitter().outer_folds[o].train_idxs)
                    )

                    outer_val_loader = provider.get_outer_val(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert len(outer_val_loader.dataset) == len(
                        set(provider._get_splitter().outer_folds[o].val_idxs)
                    )

                    outer_test_loader = provider.get_outer_test(
                        shuffle=shuffle, batch_size=batch_size
                    )

                    assert len(outer_test_loader.dataset) == len(
                        set(provider._get_splitter().outer_folds[o].test_idxs)
                    )
    rmtree('tests/data/debug/')



from typing import Tuple, List

import pytest
import torch

from pydgn.training.callback.metric import Metric, AdditiveLoss, MultiScore
from pydgn.training.event.state import State


class FakeMetric(Metric):
    def __init__(
        self,
        use_as_loss: bool = False,
        reduction: str = "mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
        )
        self.called = 0
        self.num_nodes = 20

    @property
    def name(self) -> str:
        return "Fake Metric"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        pred = torch.arange(self.num_nodes).float() + float(self.called)
        return pred, torch.zeros(1)

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        return predictions.mean()


@pytest.fixture
def fake_metric():
    def metric_init_fun(
        use_as_loss, reduction, accumulate_over_epoch, force_cpu
    ):
        return FakeMetric(
            use_as_loss, reduction, accumulate_over_epoch, force_cpu
        )

    # Return how many times the fake metric will be summed over
    return metric_init_fun


def test_metric(fake_metric):
    """
    Check that batch/epoch loss/scores are correctly computed in terms of
    averaging over batches or over the entire loss
    """
    # repetition serves to deal with additive loss, where the same mocked
    # loss will be summed over many times

    reduction = "mean"  # not necessary in this test
    for use_as_loss in [False, True]:
        for accumulate_over_epoch in [False, True]:
            for force_cpu in [False, True]:
                metric = fake_metric(
                    use_as_loss, reduction, accumulate_over_epoch, force_cpu
                )
                for num_batch_calls in [1, 10]:
                    for ep_start, ba_start, forw, comp_met, ba_end, ep_end in [
                        (
                            metric.on_training_epoch_start,
                            metric.on_training_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_training_batch_end,
                            metric.on_training_epoch_end,
                        ),
                        (
                            metric.on_eval_epoch_start,
                            metric.on_eval_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_eval_batch_end,
                            metric.on_eval_epoch_end,
                        ),
                    ]:
                        # counter used to make the score change a bit
                        metric.called = 0.0

                        num_nodes = metric.num_nodes
                        state = State(model=None, optimizer=None, device="cpu")
                        state.batch_outputs = (torch.ones(1), torch.ones(1))
                        state.batch_targets = (torch.ones(1), torch.ones(1))

                        # Simulate training epoch
                        ep_start(state)

                        for batch in range(num_batch_calls):
                            state.update(batch_loss=None)
                            state.update(batch_score=None)

                            ba_start(state)

                            forw(state)

                            comp_met(state)

                            ba_end(state)

                            # change a bit the next score
                            metric.called += 1

                        ep_end(state)

                        expected_results = [
                            torch.arange(num_nodes) + float(i)
                            for i in range(num_batch_calls)
                        ]
                        if accumulate_over_epoch:
                            expected_results = torch.cat(expected_results)
                        else:
                            # for each batch compute the average score and then
                            # average again
                            expected_results = torch.stack(
                                expected_results,
                                dim=0,
                            )
                            # average resuls for each individual batch
                            expected_results = expected_results.mean(dim=1)

                        assert (
                            expected_results.mean()
                            == state.epoch_loss[metric.name]
                            if use_as_loss
                            else state.epoch_score[metric.name]
                        )


class FakeAdditiveLoss(AdditiveLoss):
    def __init__(
        self,
        use_as_loss: bool = False,
        reduction: str = "mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **losses
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **losses
        )
        self.called = 0
        self.num_nodes = 20

    @property
    def name(self) -> str:
        return "Fake Additive Loss"


@pytest.fixture
def fake_additive_loss():
    def metric_init_fun(
        use_as_loss, reduction, accumulate_over_epoch, force_cpu
    ):
        return FakeAdditiveLoss(
            use_as_loss,
            reduction,
            accumulate_over_epoch,
            force_cpu,
            loss1="tests.training.test_metric.FakeMetric",
            loss2="tests.training.test_metric.FakeMetric",
            loss3="tests.training.test_metric.FakeMetric",
        )

    # Return how many times the fake metric will be summed over
    return metric_init_fun


def test_additive_loss(fake_additive_loss):
    """
    Check that batch/epoch loss/scores are correctly computed in terms of
    averaging over batches or over the entire loss
    """
    # repetition serves to deal with additive loss, where the same mocked
    # loss will be summed over many times

    reduction = "mean"  # not necessary in this test
    for use_as_loss in [True]:
        for accumulate_over_epoch in [False, True]:
            for force_cpu in [False, True]:
                metric = fake_additive_loss(
                    use_as_loss, reduction, accumulate_over_epoch, force_cpu
                )
                for num_batch_calls in [1, 10]:
                    for ep_start, ba_start, forw, comp_met, ba_end, ep_end in [
                        (
                            metric.on_training_epoch_start,
                            metric.on_training_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_training_batch_end,
                            metric.on_training_epoch_end,
                        ),
                        (
                            metric.on_eval_epoch_start,
                            metric.on_eval_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_eval_batch_end,
                            metric.on_eval_epoch_end,
                        ),
                    ]:
                        # counter used to make the score change a bit
                        for m in metric.losses:
                            m.called = 0.0

                        num_nodes = metric.num_nodes
                        state = State(model=None, optimizer=None, device="cpu")
                        state.batch_outputs = (torch.ones(1), torch.ones(1))
                        state.batch_targets = (torch.ones(1), torch.ones(1))

                        # Simulate training epoch
                        ep_start(state)

                        for batch in range(num_batch_calls):
                            state.update(batch_loss=None)
                            state.update(batch_score=None)

                            ba_start(state)

                            forw(state)

                            comp_met(state)

                            ba_end(state)

                            # change a bit the next score
                            for m in metric.losses:
                                m.called += 1.0

                        ep_end(state)

                        expected_results = [
                            (torch.arange(num_nodes) + float(i)) * 3.0
                            for i in range(num_batch_calls)
                        ]
                        if accumulate_over_epoch:
                            expected_results = torch.cat(expected_results)
                        else:
                            # for each batch compute the average score and then
                            # average again
                            expected_results = torch.stack(
                                expected_results,
                                dim=0,
                            )
                            # average resuls for each individual batch
                            expected_results = expected_results.mean(dim=1)

                        assert (
                            expected_results.mean()
                            == state.epoch_loss[metric.name]
                            if use_as_loss
                            else state.epoch_score[metric.name]
                        )


class FakeMultiScore(MultiScore):
    def __init__(
        self,
        use_as_loss: bool = False,
        reduction: str = "mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        main_scorer: Metric = None,
        **extra_scorers
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            main_scorer=main_scorer,
            **extra_scorers
        )
        self.called = 0
        self.num_nodes = 20

    @property
    def name(self) -> str:
        return "Fake Multi Score"


@pytest.fixture
def fake_multi_score():
    def metric_init_fun(
        use_as_loss, reduction, accumulate_over_epoch, force_cpu
    ):
        return FakeMultiScore(
            use_as_loss,
            reduction,
            accumulate_over_epoch,
            force_cpu,
            main_scorer="tests.training.test_metric.FakeMetric",
            score2="tests.training.test_metric.FakeMetric",
            score3="tests.training.test_metric.FakeMetric",
        )

    # Return how many times the fake metric will be summed over
    return metric_init_fun


def test_multi_score(fake_multi_score):
    """
    Check that batch/epoch loss/scores are correctly computed in terms of
    averaging over batches or over the entire loss
    """
    # repetition serves to deal with additive loss, where the same mocked
    # loss will be summed over many times

    reduction = "mean"  # not necessary in this test
    for use_as_loss in [False]:
        for accumulate_over_epoch in [False, True]:
            for force_cpu in [False, True]:
                metric = fake_multi_score(
                    use_as_loss, reduction, accumulate_over_epoch, force_cpu
                )
                for num_batch_calls in [1, 10]:
                    for ep_start, ba_start, forw, comp_met, ba_end, ep_end in [
                        (
                            metric.on_training_epoch_start,
                            metric.on_training_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_training_batch_end,
                            metric.on_training_epoch_end,
                        ),
                        (
                            metric.on_eval_epoch_start,
                            metric.on_eval_batch_start,
                            metric.on_forward,
                            metric.on_compute_metrics,
                            metric.on_eval_batch_end,
                            metric.on_eval_epoch_end,
                        ),
                    ]:
                        # counter used to make the score change a bit
                        for s in metric.scores:
                            s.called = 0.0

                        num_nodes = metric.num_nodes
                        state = State(model=None, optimizer=None, device="cpu")
                        state.batch_outputs = (torch.ones(1), torch.ones(1))
                        state.batch_targets = (torch.ones(1), torch.ones(1))

                        # Simulate training epoch
                        ep_start(state)

                        for batch in range(num_batch_calls):
                            state.update(batch_loss=None)
                            state.update(batch_score=None)

                            ba_start(state)

                            forw(state)

                            comp_met(state)

                            ba_end(state)

                            # change a bit the next score
                            for s in metric.scores:
                                s.called += 1.0

                        ep_end(state)

                        expected_results = [
                            torch.arange(num_nodes) + float(i)
                            for i in range(num_batch_calls)
                        ]
                        if accumulate_over_epoch:
                            expected_results = torch.cat(expected_results)
                        else:
                            # for each batch compute the average score and then
                            # average again
                            expected_results = torch.stack(
                                expected_results,
                                dim=0,
                            )
                            # average resuls for each individual batch
                            expected_results = expected_results.mean(dim=1)

                        assert (
                            expected_results.mean()
                            == state.epoch_loss[metric.get_main_metric_name()]
                            if use_as_loss
                            else state.epoch_score[
                                metric.get_main_metric_name()
                            ]
                        )



from pydoc import locate
from typing import Callable


def s2c(class_name: str) -> Callable[..., object]:
    r"""
    Converts a dotted path to the corresponding class

    Args:
         class_name (str): dotted path to class name

    Returns:
        the class to be used
    """
    result = locate(class_name)
    if result is None:
        raise ImportError(
            f"The (dotted) path '{class_name}' is unknown. "
            f"Check your configuration."
        )
    return result



import random
from typing import Callable, Tuple, List

import numpy as np
import torch

from pydgn.data.provider import DataProvider
from pydgn.evaluation.config import Config
from pydgn.evaluation.util import return_class_and_args
from pydgn.experiment.util import s2c
from pydgn.log.logger import Logger
from pydgn.model.interface import ModelInterface, ReadoutInterface
from pydgn.static import DEFAULT_ENGINE_CALLBACK
from pydgn.training.engine import TrainingEngine


class Experiment:
    r"""
    Class that handles a single experiment.

    Args:
        model_configuration (dict): the dictionary holding the
            experiment-specific configuration
        exp_path (str): path to the experiment folder
        exp_seed (int): the experiment's seed to use
    """

    def __init__(
        self, model_configuration: dict, exp_path: str, exp_seed: int
    ):
        self.model_config = Config(model_configuration)
        self.exp_path = exp_path
        self.exp_seed = exp_seed
        # Set seed here to aid reproducibility
        np.random.seed(self.exp_seed)
        torch.manual_seed(self.exp_seed)
        torch.cuda.manual_seed(self.exp_seed)
        random.seed(self.exp_seed)

        # torch.use_deterministic_algorithms(True) for future versions of
        # Pytorch
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    def _return_class_and_args(
        config: Config, key: str
    ) -> Tuple[Callable[..., object], dict]:
        r"""
        Returns the class and arguments associated to a specific key in the
        configuration file.

        Args:
            config: the configuration dictionary
            key: a string representing a particular class in the
                configuration dictionary

        Returns:
            a tuple (class, dict of arguments), or (None, None) if the key
            is not present in the config dictionary
        """
        if key not in config or config[key] is None:
            return None, None
        elif isinstance(config[key], str):
            return s2c(config[key]), {}
        elif isinstance(config[key], dict):
            return (
                s2c(config[key]["class_name"]),
                config[key]["args"] if "args" in config[key] else {},
            )
        else:
            raise NotImplementedError(
                "Parameter has not been formatted properly"
            )

    def _create_model(
        self,
        dim_node_features: int,
        dim_edge_features: int,
        dim_target: int,
        readout_classname: str,
        config: Config,
    ) -> ModelInterface:
        r"""
        Instantiates a model that implements the
        :class:`~pydgn.model.model.ModelInterface` interface

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension
            readout_classname (str): string containing the model's class
            config (:class:`~pydgn.evaluation.config.Config`):
                the configuration dictionary

        Returns:
            a model that implements the
            :class:`~pydgn.model.model.ModelInterface` interface
        """
        model = s2c(config["model"])(
            dim_node_features=dim_node_features,
            dim_edge_features=dim_edge_features,
            dim_target=dim_target,
            readout_class=s2c(readout_classname)
            if readout_classname is not None
            else None,
            config=config,
        )

        # move to device
        # model .to() may not return anything
        model.to(self.model_config.device)
        return model

    def create_supervised_model(
        self, dim_node_features: int, dim_edge_features: int, dim_target: int
    ) -> ModelInterface:
        r"""
        Instantiates a **supervised** model that implements the
        :class:`~pydgn.model.model.ModelInterface` interface,
        using the ``supervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the
            :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = (
            self.model_config.supervised_config["readout"]
            if "readout" in self.model_config.supervised_config
            else None
        )
        return self._create_model(
            dim_node_features,
            dim_edge_features,
            dim_target,
            readout_classname,
            self.model_config.supervised_config,
        )

    def create_supervised_readout(
        self, dim_node_features: int, dim_edge_features: int, dim_target: int
    ) -> ReadoutInterface:
        r"""
        Instantiates an **supervised** readout that implements the
        :class:`~pydgn.model.model.ReadoutInterface` interface,
        using the ``supervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the
            :class:`~pydgn.model.model.ReadoutInterface` interface
        """
        return s2c(self.model_config.supervised_config["readout"])(
            dim_node_features=dim_node_features,
            dim_edge_features=dim_edge_features,
            dim_target=dim_target,
            config=self.model_config.supervised_config,
        )

    def create_unsupervised_model(
        self, dim_node_features: int, dim_edge_features: int, dim_target: int
    ) -> ModelInterface:
        r"""
        Instantiates an **unsupervised** model that implements the
        :class:`~pydgn.model.model.ModelInterface` interface,
        using the ``unsupervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the
            :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = (
            self.model_config.unsupervised_config["readout"]
            if "readout" in self.model_config.unsupervised_config
            else None
        )
        return self._create_model(
            dim_node_features,
            dim_edge_features,
            dim_target,
            readout_classname,
            self.model_config.unsupervised_config,
        )

    def create_incremental_model(
        self,
        dim_node_features: (int),
        dim_edge_features: (int),
        dim_target: (int),
        depth: (int),
        prev_outputs_to_consider: List[int],
    ) -> ModelInterface:
        r"""
        Instantiates a layer of an incremental architecture.
        It assumes the config file has a field ``layer_config``
        and another ``layer_config.arbitrary_function_config``
        that holds any kind of information for the arbitrary
        function of an incremental architecture

        Args:
            dim_node_features: input node features
            dim_edge_features: input edge features
            dim_target: target size
            depth: current depth of the architecture
            prev_outputs_to_consider: A list of previous layers to consider,
                e.g., [1,2] means the last two previous layers.

        Returns:
            a layer of a model that implements the
            :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = self.model_config.layer_config.get("readout", None)
        self.model_config.layer_config["depth"] = depth
        self.model_config.layer_config[
            "prev_outputs_to_consider"
        ] = prev_outputs_to_consider
        return self._create_model(
            dim_node_features,
            dim_edge_features,
            dim_target,
            readout_classname,
            self.model_config.layer_config,
        )

    def _create_engine(
        self,
        config: Config,
        model: ModelInterface,
        device: str,
        evaluate_every: int,
        reset_eval_model_hidden_state: bool,
    ) -> TrainingEngine:
        r"""
        Utility that instantiates the training engine. It looks for
        pre-defined fields in the configuration file, i.e. ``loss``,
        ``scorer``, ``optimizer``, ``scheduler``, ``gradient_clipper``,
        ``early_stopper`` and ``plotter``, all of which should be classes
        implementing the :class:`~pydgn.training.event.handler.EventHandler`
        interface

        Args:
            config (:class:`~pydgn.evaluation.config.Config`):
                the configuration dictionary
            model: the  model that needs be trained
            device (str): the string with the CUDA device to be used,
                or ``cpu``
            evaluate_every (int): number of epochs after which to
                log information
            reset_eval_model_hidden_state (bool): [temporal graph learning]
                Used when we want to reset the state after performing
                previous inference. It should be ``False`` when we are
                dealing with a single temporal graph sequence,
                because we don't want to reset the hidden state after
                processing the previous [training/validation] time steps.

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """

        loss_class, loss_args = return_class_and_args(config, "loss")
        loss_args.update(device=device)
        loss = (
            loss_class(use_as_loss=True, **loss_args)
            if loss_class is not None
            else None
        )

        scorer_class, scorer_args = return_class_and_args(config, "scorer")
        scorer_args.update(device=device)
        scorer = (
            scorer_class(use_as_loss=False, **scorer_args)
            if scorer_class is not None
            else None
        )

        optim_class, optim_args = return_class_and_args(config, "optimizer")
        optimizer = (
            optim_class(model=model, **optim_args)
            if optim_class is not None
            else None
        )

        sched_class, sched_args = return_class_and_args(config, "scheduler")
        if sched_args is not None:
            sched_args["optimizer"] = optimizer.optimizer
        scheduler = (
            sched_class(**sched_args) if sched_class is not None else None
        )
        # Remove the optimizer obj ow troubles when dumping the config file
        if sched_args is not None:
            sched_args.pop("optimizer", None)

        grad_clip_class, grad_clip_args = return_class_and_args(
            config, "gradient_clipper"
        )
        grad_clipper = (
            grad_clip_class(**grad_clip_args)
            if grad_clip_class is not None
            else None
        )

        early_stop_class, early_stop_args = return_class_and_args(
            config, "early_stopper"
        )
        early_stopper = (
            early_stop_class(**early_stop_args)
            if early_stop_class is not None
            else None
        )

        plot_class, plot_args = return_class_and_args(config, "plotter")
        plotter = (
            plot_class(exp_path=self.exp_path, **plot_args)
            if plot_class is not None
            else None
        )

        store_last_checkpoint = config.get("checkpoint", False)
        engine_class, engine_args = return_class_and_args(config, "engine")
        engine_callback = s2c(
            engine_args.get("engine_callback", DEFAULT_ENGINE_CALLBACK)
        )

        engine = engine_class(
            engine_callback=engine_callback,
            model=model,
            loss=loss,
            optimizer=optimizer,
            scorer=scorer,
            scheduler=scheduler,
            early_stopper=early_stopper,
            gradient_clipper=grad_clipper,
            device=device,
            plotter=plotter,
            exp_path=self.exp_path,
            evaluate_every=evaluate_every,
            store_last_checkpoint=store_last_checkpoint,
            reset_eval_model_hidden_state=reset_eval_model_hidden_state,
        )
        return engine

    def create_supervised_engine(
        self, model: ModelInterface
    ) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``supervised_config``
        key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get(
            "reset_eval_model_hidden_state", True
        )

        return self._create_engine(
            self.model_config.supervised_config,
            model,
            device,
            evaluate_every,
            reset_eval_model_hidden_state,
        )

    def create_unsupervised_engine(
        self, model: ModelInterface
    ) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``unsupervised_config``
        key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get(
            "reset_eval_model_hidden_state", True
        )
        return self._create_engine(
            self.model_config.unsupervised_config,
            model,
            device,
            evaluate_every,
            reset_eval_model_hidden_state,
        )

    def create_incremental_engine(
        self, model: ModelInterface
    ) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``layer_config``
        key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get(
            "reset_eval_model_hidden_state", True
        )
        return self._create_engine(
            self.model_config.layer_config,
            model,
            device,
            evaluate_every,
            reset_eval_model_hidden_state,
        )

    def run_valid(self, dataset_getter, logger) -> Tuple[dict, dict]:
        r"""
        This function returns the training and validation results for a
        `model selection run`.
        **Do not attempt to load the test set inside this method!**
        **If possible, rely on already available subclasses of this class**.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`):
                a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training and test dictionaries.
            Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself
            with other fields to be used by the evaluator.
        """
        raise NotImplementedError("You must implement this function!")

    def run_test(
        self, dataset_getter: DataProvider, logger: Logger
    ) -> Tuple[dict, dict, dict]:
        """
        This function returns the training, validation and test results
        for a `final run`.
        **Do not use the test to train the model nor for
        early stopping reasons!**
        **If possible, rely on already available subclasses of this class**.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`):
                a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training,validation,test dictionaries.
            Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself
            with other fields to be used by the evaluator.
        """
        raise NotImplementedError("You must implement this function!")



import json
from shutil import rmtree

import numpy as np
import yaml

from pydgn.evaluation.evaluator import RiskAssesser
from pydgn.evaluation.grid import Grid
from pydgn.experiment.experiment import Experiment
from pydgn.static import DATA_SPLITS_FILE, LOSS, SCORE, MAIN_LOSS, MAIN_SCORE


class FakeTask(Experiment):
    def run_valid(self, dataset_getter, logger):
        outer_k = dataset_getter.outer_k
        inner_k = dataset_getter.inner_k

        train_loss = {MAIN_LOSS: outer_k + inner_k}
        train_score = {MAIN_SCORE: outer_k + inner_k}
        val_loss = {MAIN_LOSS: outer_k + inner_k + 1}
        val_score = {MAIN_SCORE: outer_k + inner_k + 1}

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}

        return train_res, val_res

    def run_test(self, dataset_getter, logger):
        outer_k = dataset_getter.outer_k

        train_loss = {MAIN_LOSS: outer_k}
        train_score = {MAIN_SCORE: outer_k}
        val_loss = {MAIN_LOSS: outer_k + 1}
        val_score = {MAIN_SCORE: outer_k + 1}
        test_loss = {MAIN_LOSS: outer_k + 2}
        test_score = {MAIN_SCORE: outer_k + 2}

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        test_res = {LOSS: test_loss, SCORE: test_score}

        return train_res, val_res, test_res


# This test activates most of the library's main routines.
def test_evaluator():
    results_folder = "tests/evaluation/debug_evaluator/"
    search = Grid(
        yaml.load(
            open("tests/evaluation/grid_search.yml", "r"),
            Loader=yaml.FullLoader,
        )
    )
    search.telegram_config = None
    splits_filepath = search.configs_dict.get(DATA_SPLITS_FILE)

    evaluator = RiskAssesser(
        10,
        10,
        FakeTask,
        results_folder,
        splits_filepath,
        search,
        10,
        True,
        0,
        base_seed=42,
    )
    # the goal is to evaluate the code that computes results, to ensure
    # it is correct. For this, we don't need debug=False, ow Ray should be
    # started.
    evaluator.risk_assessment(debug=True)

    for outer_k in range(10):
        inner_train_results = np.array([float(i + outer_k) for i in range(10)])
        inner_val_results = np.array(
            [float(i + outer_k) + 1 for i in range(10)]
        )

        ms_results = json.load(
            open(
                "tests/evaluation/debug_evaluator/"
                + f"MODEL_ASSESSMENT/OUTER_FOLD_{outer_k+1}/"
                + "MODEL_SELECTION/winner_config.json",
                "r",
            )
        )

        assert ms_results["avg_training_loss"] == inner_train_results.mean()
        assert ms_results["avg_validation_loss"] == inner_val_results.mean()

    outer_train_results = np.array([float(i) for i in range(10)])
    outer_val_results = np.array([float(i) + 1 for i in range(10)])
    outer_test_results = np.array([float(i) + 2 for i in range(10)])

    ass_results = json.load(
        open(
            "tests/evaluation/debug_evaluator/"
            + "MODEL_ASSESSMENT/assessment_results.json",
            "r",
        )
    )

    assert ass_results["avg_training_main_score"] == outer_train_results.mean()
    assert ass_results["avg_validation_main_score"] == outer_val_results.mean()
    assert ass_results["avg_test_main_score"] == outer_test_results.mean()

    assert ass_results["std_training_main_score"] == outer_train_results.std()
    assert ass_results["std_validation_main_score"] == outer_val_results.std()
    assert ass_results["std_test_main_score"] == outer_test_results.std()

    rmtree(results_folder)



from shutil import rmtree

import pytest
import yaml

from pydgn.data.util import preprocess_data
from pydgn.launch_experiment import evaluation
from pydgn.static import DEBUG, DEVICE, CONFIG_FILE


def test_dataset_creation():
    config = yaml.load(
        open("tests/integration/graph_class_dataset_config.yml", "r"),
        Loader=yaml.FullLoader,
    )
    preprocess_data(config)


@pytest.mark.dependency(depends=["test_dataset_creation"])
def test_supervised_experiment():
    class MockConfig:
        def __init__(self, d):
            for key in d.keys():
                setattr(self, key, d[key])

    config = {}
    config[CONFIG_FILE] = "tests/integration/sup_exp_config.yml"
    config[DEBUG] = True
    config = MockConfig(config)
    evaluation(config)


@pytest.mark.dependency(depends=["test_dataset_creation"])
def test_semi_supervised_experiment():
    class MockConfig:
        def __init__(self, d):
            for key in d.keys():
                setattr(self, key, d[key])

    config = {}
    config[CONFIG_FILE] = "tests/integration/semi_sup_exp_config.yml"
    config[DEBUG] = True
    config = MockConfig(config)
    evaluation(config)


@pytest.mark.dependency(
    depends=["test_supervised_experiment", "test_semi_supervised_experiment"]
)
def test_cleanup():
    rmtree("tests/integration/debug")



import inspect
import os
import os.path as osp
import warnings
from typing import Optional, Callable

import torch
from torch_geometric.transforms import Compose

from pydgn.data.dataset import DatasetInterface
from pydgn.experiment.util import s2c


def get_or_create_dir(path: str) -> str:
    r"""
    Creates directories associated to the specified path if they are missing,
    and it returns the path string.

    Args:
        path (str): the path

    Returns:
        the same path as the given argument
    """
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def check_argument(cls: object, arg_name: str) -> bool:
    r"""
    Checks whether ``arg_name`` is in the signature of a method or class.

    Args:
        cls (object): the class to inspect
        arg_name (str): the name to look for

    Returns:
        ``True`` if the name was found, ``False`` otherwise
    """
    sign = inspect.signature(cls)
    return arg_name in sign.parameters.keys()


def filter_adj(
    edge_index: torch.Tensor, edge_attr: torch.Tensor, mask: torch.Tensor
) -> (torch.Tensor, Optional[torch.Tensor]):
    r"""
    Adapted from pytorch-geometric. Does the same thing but with
    a different signature

    Args:
        edge_index (torch.Tensor): the usual PyG matrix of edge indices
        edge_attr (torch.Tensor): the usual PyG matrix of edge attributes
        mask (torch.Tensor): boolean tensor with edges to filter

    Returns:
        a tuple (filtered edge index, filtered edge attr or ``None``
        if ``edge_attr`` is ``None``)
    """
    row, col = edge_index
    filtered_edge_index = row[mask], col[mask]
    return filtered_edge_index, None if edge_attr is None else edge_attr[mask]


def preprocess_data(options: dict):
    r"""
    One of the main functions of the PyDGN library. Used to create the dataset
    and its associated files that ensure the correct functioning of the
    data loading steps.

    Args:
        options (dict): a dictionary of dataset/splitter arguments as
            defined in the data configuration file used.

    """
    data_info = options.pop("dataset")
    if "class_name" not in data_info:
        raise ValueError("You must specify 'class_name' in your dataset.")
    dataset_class = s2c(data_info.pop("class_name"))
    dataset_args = data_info.pop("args")
    data_root = data_info.pop("root")

    ################################

    # more experimental stuff here

    dataset_kwargs = data_info.pop("other_args", {})

    pre_transforms = None
    pre_transforms_opt = data_info.pop("pre_transform", None)
    if pre_transforms_opt is not None:
        pre_transforms = []
        for pre_transform in pre_transforms_opt:
            pre_transform_class = s2c(pre_transform["class_name"])
            args = pre_transform.pop("args", {})
            pre_transforms.append(pre_transform_class(**args))
        dataset_kwargs.update(pre_transform=Compose(pre_transforms))

    pre_filters = None
    pre_filters_opt = data_info.pop("pre_filter", None)
    if pre_filters_opt is not None and check_argument(
        dataset_class, "pre_filter"
    ):
        pre_filters = []
        for pre_filter in pre_filters_opt:
            pre_filter_class = s2c(pre_filter["class_name"])
            args = pre_filter.pop("args", {})
            pre_filters.append(pre_filter_class(**args))
        dataset_kwargs.update(pre_filter=Compose(pre_filters))

    transforms_opt = data_info.pop("transform", None)

    if transforms_opt is not None:
        transforms = []
        for transform in transforms_opt:
            transform_class = s2c(transform["class_name"])
            args = transform.pop("args", {})
            transforms.append(transform_class(**args))
        dataset_kwargs.update(transform=Compose(transforms))

    dataset_args.update(dataset_kwargs)

    ################################

    dataset = dataset_class(**dataset_args)
    assert hasattr(
        dataset, "name"
    ), "Dataset instance should have a name attribute!"

    # Store dataset additional arguments in a separate file
    kwargs_folder = osp.join(data_root, dataset.name)
    kwargs_path = osp.join(kwargs_folder, "dataset_kwargs.pt")

    get_or_create_dir(kwargs_folder)
    torch.save(dataset_args, kwargs_path)

    # Process data splits

    splits_info = options.pop("splitter")
    splits_root = splits_info.pop("root")
    if "class_name" not in splits_info:
        raise ValueError("You must specify 'class_name' in your splitter.")
    splitter_class = s2c(splits_info.pop("class_name"))
    splitter_args = splits_info.pop("args")
    splitter = splitter_class(**splitter_args)

    splits_dir = get_or_create_dir(osp.join(splits_root, dataset.name))
    splits_path = osp.join(
        splits_dir,
        f"{dataset.name}_outer{splitter.n_outer_folds}"
        f"_inner{splitter.n_inner_folds}.splits",
    )

    if not os.path.exists(splits_path):
        has_targets, targets = splitter.get_targets(dataset)
        # The splitter is in charge of eventual stratifications
        splitter.split(dataset, targets=targets if has_targets else None)
        splitter.save(splits_path)
    else:
        print("Data splits are already present, I will not overwrite them.")


def load_dataset(
    data_root: str,
    dataset_name: str,
    dataset_class: Callable[..., DatasetInterface],
    **kwargs: dict,
) -> DatasetInterface:
    r"""
    Loads the dataset using the ``dataset_kwargs.pt`` file created when parsing
    the data config file.

    Args:
        data_root (str): path of the folder that contains the dataset folder
        dataset_name (str): name of the dataset (same as the name of the
            dataset folder that has been already created)
        dataset_class
            (Callable[..., :class:`~pydgn.data.dataset.DatasetInterface`]):
            the class of the dataset to instantiate with the parameters
            stored in the ``dataset_kwargs.pt`` file.
        kwargs (dict): additional arguments to be passed to the
            dataset (potentially provided by a DataProvider)

    Returns:
        a :class:`~pydgn.data.dataset.DatasetInterface` object
    """
    # Load arguments
    kwargs_path = osp.join(data_root, dataset_name, "dataset_kwargs.pt")
    if not os.path.exists(kwargs_path):  # backward compatibility
        kwargs_path = osp.join(
            data_root, dataset_name, "processed", "dataset_kwargs.pt"
        )

    dataset_args = torch.load(kwargs_path)

    # Overwrite original data_root field, which may have changed
    dataset_args["root"] = data_root

    # pass extra arguments to dataset
    dataset_args.update(kwargs)

    with warnings.catch_warnings():
        # suppress PyG warnings
        warnings.simplefilter("ignore")
        dataset = dataset_class(**dataset_args)

    return dataset


def to_lower_triangular(edge_index: torch.Tensor):
    r"""
    Transform Pytorch Geometric undirected edge index into its
    "lower triangular counterpart"
    """
    row, col = edge_index
    lower_tri_mask = row > col
    row, col = row[lower_tri_mask], col[lower_tri_mask]
    lower_tri_edge_index = torch.cat(
        (row.unsqueeze(0), col.unsqueeze(0)), dim=0
    )
    return lower_tri_edge_index



import torch
from torch_geometric.utils import degree


class ConstantIfEmpty:
    r"""
    Adds a constant value to each node feature only if x is None.

    Args:
        value (int): The value to add. Default is ``1``
    """

    def __init__(self, value=1):
        self.value = value

    def __call__(self, data):
        """
        Transforms the data object by adding a constant value as the sole node
        feature (if none are present).
        """
        if data.x is None:
            c = torch.full((data.num_nodes, 1), self.value, dtype=torch.float)
            data.x = c
        return data

    def __repr__(self):
        """
        String representation of the transform.
        """
        return "{}(value={})".format(self.__class__.__name__, self.value)


class ConstantEdgeIfEmpty:
    r"""
    Adds a constant value to each edge feature only if edge_attr is None.

    Args:
        value (int): The value to add. Default is ``1``)
    """

    def __init__(self, value=1):
        self.value = value

    def __call__(self, data):
        """
        Transforms the data object by adding a constant value as the sole edge
        feature (if none are present).
        """
        if data.edge_attr is None:
            c = torch.full(
                (data.edge_index.shape[1], 1), self.value, dtype=torch.float
            )
            data.edge_attr = c
        return data

    def __repr__(self):
        """
        String representation of the transform.
        """
        return "{}(value={})".format(self.__class__.__name__, self.value)


class Degree:
    r"""
    Adds the node degree to the node features.

    Args:
        in_degree (bool): If set to :obj:`True`, will compute the in-degree of
            nodes instead of the out-degree.
        Not relevant if the graph is undirected (default: :obj:`False`).
        cat (bool): Concat node degrees to node features instead of replacing
            them. (default: :obj:`True`)
    """

    def __init__(self, in_degree: bool = False, cat: bool = True):
        self.in_degree = in_degree
        self.cat = cat

    def __call__(self, data):
        """
        Transforms the data object by adding the in or out-degree of each
        node as a feature.
        """
        idx, x = data.edge_index[1 if self.in_degree else 0], data.x
        deg = degree(idx, data.num_nodes, dtype=torch.float).view(-1, 1)

        if x is not None and self.cat:
            data.x = torch.cat([x, deg.to(x.dtype)], dim=-1)
        else:
            data.x = deg

        return data

    def __repr__(self):
        """
        String representation of the transform.
        """
        return "{}".format(self.__class__.__name__)



from typing import Callable, Tuple, Optional, List

import torch
from torch_geometric.data import Batch


class ModelInterface(torch.nn.Module):
    r"""
    Provides the signature for any main model to be trained under PyDGN

    Args:
        dim_node_features (int): dimension of node features
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        dim_edge_features (int): dimension of edge features
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        dim_target (int): dimension of the target
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        readout_class (Callable[...,:class:`torch.nn.Module`]):
            class of the module implementing the readout. This is optional,
            but useful to put different readouts to try in the config file
        config (dict): config dictionary containing all the necessary
            hyper-parameters plus additional information (if needed)
    """

    def __init__(
        self,
        dim_node_features: int,
        dim_edge_features: int,
        dim_target: int,
        readout_class: Callable[..., torch.nn.Module],
        config: dict,
    ):
        super().__init__()
        self.dim_node_features = dim_node_features
        self.dim_edge_features = dim_edge_features
        self.dim_target = dim_target
        self.readout_class = readout_class
        self.config = config

    def forward(
        self, data: Batch
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        r"""
        Performs a forward pass over a batch of graphs

        Args:
            data (:class:`torch_geometric.data.Batch`): a batch of graphs

        Returns:
            a tuple (model's output, [optional] node embeddings,
            [optional] additional outputs
        """
        raise NotImplementedError("You need to implement this method!")


class ReadoutInterface(torch.nn.Module):
    r"""
    Provides the signature for any readout to be trained under PyDGN

    Args:
        dim_node_features (int): dimension of node features
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        dim_edge_features (int): dimension of edge features
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        dim_target (int): dimension of the target
            (according to the :class:`~pydgn.data.dataset.DatasetInterface`
            property)
        config (dict): config dictionary containing all the necessary
            hyper-parameters plus additional information (if needed)
    """

    def __init__(
        self,
        dim_node_features: int,
        dim_edge_features: int,
        dim_target: int,
        config: dict,
    ):
        super().__init__()
        self.dim_node_features = dim_node_features
        self.dim_edge_features = dim_edge_features
        self.dim_target = dim_target
        self.config = config

    def forward(
        self, node_embeddings: torch.tensor, batch: torch.Tensor, **kwargs
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        r"""
        Performs a forward pass over a batch of graphs

        Args:
            node_embeddings (:class:`torch_geometric.data.Batch`):
                the node embeddings
            batch (:class:`torch.Tensor`): the usual ``batch`` object of PyG
            kwargs (dict): additional and optional arguments

        Returns:
            a tuple (model's output, [optional] node embeddings,
            [optional] additional outputs
        """
        raise NotImplementedError("You need to implement this method!")



import copy
import os
from pathlib import Path

from pydgn.static import *
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State
from pydgn.training.util import atomic_save


class EngineCallback(EventHandler):
    r"""
    Class responsible for fetching data and handling current-epoch checkpoints
     at training time.

    Args:
        store_last_checkpoint (bool): if ``True``, keep the model's
            checkpoint for the last training epoch
    """

    def __init__(self, store_last_checkpoint: bool):
        super().__init__()
        self.store_last_checkpoint = store_last_checkpoint

    # Allows to profile data loading
    def on_fetch_data(self, state: State):
        """
        Fetches next batch of data from loader and updates the `batch_input`
        field of the state

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information
        """
        data = state.loader_iterable.next()
        state.update(batch_input=data)

    def on_forward(self, state: State):
        """
        Calls the forward method of the model and stores the outputs in the
        `batch_outputs` field of the state.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        # Forward pass
        outputs = state.model.forward(state.batch_input)
        state.update(batch_outputs=outputs)

    def on_epoch_end(self, state: State):
        """
        Stores the checkpoint in a dictionary with the following fields:

        * ``EPOCH`` (as defined in ``pydgn.static``)
        * ``MODEL_STATE`` (as defined in ``pydgn.static``)
        * ``OPTIMIZER_STATE`` (as defined in ``pydgn.static``)
        * ``SCHEDULER_STATE`` (as defined in ``pydgn.static``)
        * ``STOP_TRAINING`` (as defined in ``pydgn.static``)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        # Save last checkpoint
        if self.store_last_checkpoint:
            if not os.path.exists(Path(state.exp_path)):
                os.makedirs(Path(state.exp_path))

            last_ckpt = {
                EPOCH: state.epoch,
                MODEL_STATE: copy.deepcopy(state.model.state_dict()),
                OPTIMIZER_STATE: getattr(state, OPTIMIZER_STATE, None),
                SCHEDULER_STATE: getattr(state, SCHEDULER_STATE, None),
                STOP_TRAINING: state.stop_training,
            }
            last_ckpt.update(state.epoch_results)
            atomic_save(
                last_ckpt, Path(state.exp_path, LAST_CHECKPOINT_FILENAME)
            )


class IterableEngineCallback(EngineCallback):
    r"""
    Class that extends :class:`pydgn.training.callback.EngineCallback`
    to the processing of Iterable-style datasets.
    Needs to be used together with the appropriate engine class.
    """

    def on_fetch_data(self, state: State):
        """
        Fetches next batch of data from loader (if any, as data comes
        from a stream of unknown length)
        and updates the `batch_input` field of the state

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        try:
            data = next(state.loader_iterable)
            state.update(batch_input=data)
        except StopIteration as e:
            state.update(stop_fetching=True)
            state.update(batch_input=None)


class TemporalEngineCallback(EngineCallback):
    r"""
    Class that extends :class:`pydgn.training.callback.EngineCallback`
    to the processing of temporal datasets.
    Needs to be used together with the appropriate engine class.
    """

    def on_forward(self, state):
        """
        Calls the forward method of the model and stores the outputs in the
        `batch_outputs` field of the state. In addition to the input, passes
        to the model the hidden state computed at the previous time step.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        # Forward pass, the last hidden state gets passed as additional
        # argument
        outputs = state.model.forward(
            state.batch_input, state.last_hidden_state
        )
        state.update(batch_outputs=outputs)



import datetime
import math
import os
import random
from typing import Tuple, Callable

import tqdm

from pydgn.experiment.util import s2c
from pydgn.static import *


def return_class_and_args(
    config: dict, key: str, return_class_name: bool = False
) -> Tuple[Callable[..., object], dict]:
    r"""
    Returns the class and arguments associated to a specific key in the
    configuration file.

    Args:
        config (dict): the configuration dictionary
        key (str): a string representing a particular class in the
            configuration dictionary
        return_class_name (bool): if ``True``, returns the class name as a
            string rather than the class object

    Returns:
        a tuple (class, dict of arguments), or (None, None) if the key
        is not present in the config dictionary
    """
    if key not in config or config[key] is None:
        return None, None
    elif isinstance(config[key], str):
        return s2c(config[key]), {}
    elif isinstance(config[key], dict):
        return (
            s2c(config[key]["class_name"])
            if not return_class_name
            else config[key]["class_name"],
            config[key]["args"] if "args" in config[key] else {},
        )
    else:
        raise NotImplementedError(
            f"Parameter {key} " f"has not been formatted properly"
        )


def clear_screen():
    """
    Clears the CLI interface.
    """
    try:
        os.system("clear")
    except Exception as e:
        try:
            os.system("cls")
        except Exception:
            pass


class ProgressManager:
    r"""
    Class that is responsible for drawing progress bars.

    Args:
        outer_folds (int): number of external folds for model assessment
        inner_folds (int): number of internal folds for model selection
        no_configs (int): number of possible configurations in model selection
        final_runs (int): number of final runs per outer fold once the
            best model has been selected
        show (bool): whether to show the progress bar or not.
            Default is ``True``
    """

    # Possible vars of ``bar_format``:
    #       * ``l_bar, bar, r_bar``,
    #       * ``n, n_fmt, total, total_fmt``,
    #       * ``percentage, elapsed, elapsed_s``,
    #       * ``ncols, nrows, desc, unit``,
    #       * ``rate, rate_fmt, rate_noinv``,
    #       * ``rate_noinv_fmt, rate_inv, rate_inv_fmt``,
    #       * ``postfix, unit_divisor, remaining, remaining_s``

    def __init__(
        self, outer_folds, inner_folds, no_configs, final_runs, show=True
    ):
        self.ncols = 100
        self.outer_folds = outer_folds
        self.inner_folds = inner_folds
        self.no_configs = no_configs
        self.final_runs = final_runs
        self.pbars = []
        self.show = show

        if not self.show:
            return

        clear_screen()
        self.show_header()
        for i in range(self.outer_folds):
            for j in range(self.inner_folds):
                self.pbars.append(self._init_selection_pbar(i, j))

        for i in range(self.outer_folds):
            self.pbars.append(self._init_assessment_pbar(i))

        self.show_footer()

        self.times = [{} for _ in range(len(self.pbars))]

    def _init_selection_pbar(self, i: int, j: int):
        """
        Initializes the progress bar for model selection

        Args:
            i (int): the id of the outer fold (from 0 to outer folds - 1)
            j (int): the id of the inner fold (from 0 to inner folds - 1)
        """
        position = i * self.inner_folds + j
        pbar = tqdm.tqdm(
            total=self.no_configs,
            ncols=self.ncols,
            ascii=True,
            position=position,
            unit="config",
            bar_format=" {desc} {percentage:3.0f}%|"
            "{bar}|{n_fmt}/{total_fmt}{postfix}",
        )
        pbar.set_description(f"Out_{i + 1}/Inn_{j + 1}")
        mean = str(datetime.timedelta(seconds=0))
        pbar.set_postfix_str(f"(1 cfg every {mean})")
        return pbar

    def _init_assessment_pbar(self, i: int):
        """
        Initializes the progress bar for risk assessment

        Args:
            i (int): the id of the outer fold (from 0 to outer folds - 1)
        """
        position = self.outer_folds * self.inner_folds + i
        pbar = tqdm.tqdm(
            total=self.final_runs,
            ncols=self.ncols,
            ascii=True,
            position=position,
            unit="config",
            bar_format=" {desc} {percentage:3.0f}%|"
            "{bar}|{n_fmt}/{total_fmt}{postfix}",
        )
        pbar.set_description(f"Final run {i + 1}")
        mean = str(datetime.timedelta(seconds=0))
        pbar.set_postfix_str(f"(1 run every {mean})")
        return pbar

    def show_header(self):
        """
        Prints the header of the progress bar
        """
        """
        \033[F --> move cursor to the beginning of the previous line
        \033[A --> move cursor up one line
        \033[<N>A --> move cursor up N lines
        """
        print(
            f'\033[F\033[A{"*" * ((self.ncols - 21) // 2 + 1)} '
            f'Experiment Progress {"*" * ((self.ncols - 21) // 2)}\n'
        )

    def show_footer(self):
        """
        Prints the footer of the progress bar
        """
        pass  # need to work how how to print after tqdm

    def refresh(self):
        """
        Refreshes the progress bar
        """

        self.show_header()
        for i, pbar in enumerate(self.pbars):

            # When resuming, do not consider completed exp. (delta approx. < 1)
            completion_times = [
                delta
                for k, (delta, completed) in self.times[i].items()
                if completed and delta > 1
            ]

            if len(completion_times) > 0:
                min_seconds = min(completion_times)
                max_seconds = max(completion_times)
                mean_seconds = sum(completion_times) / len(completion_times)
            else:
                min_seconds = 0
                max_seconds = 0
                mean_seconds = 0

            mean_time = str(datetime.timedelta(seconds=mean_seconds)).split(
                "."
            )[0]
            min_time = str(datetime.timedelta(seconds=min_seconds)).split(".")[
                0
            ]
            max_time = str(datetime.timedelta(seconds=max_seconds)).split(".")[
                0
            ]

            pbar.set_postfix_str(
                f"min:{min_time}|avg:{mean_time}|max:{max_time}"
            )

            pbar.refresh()
        self.show_footer()

    def update_state(self, msg: dict):
        """
        Updates the state of the progress bar (different from showing it
        on screen, see :func:`refresh`) once a message is received

        Args:
            msg (dict): message with updates to be parsed
        """
        if not self.show:
            return

        try:
            type = msg.get("type")

            if type == END_CONFIG:
                outer_fold = msg.get(OUTER_FOLD)
                inner_fold = msg.get(INNER_FOLD)
                config_id = msg.get(CONFIG_ID)
                position = outer_fold * self.inner_folds + inner_fold
                elapsed = msg.get(ELAPSED)
                configs_times = self.times[position]
                # Compute delta t for a specific config
                configs_times[config_id] = (
                    elapsed,
                    True,
                )  # (time.time() - configs_times[config_id][0], True)
                # Update progress bar
                self.pbars[position].update()
                self.refresh()
            elif type == END_FINAL_RUN:
                outer_fold = msg.get(OUTER_FOLD)
                run_id = msg.get(RUN_ID)
                position = self.outer_folds * self.inner_folds + outer_fold
                elapsed = msg.get(ELAPSED)
                configs_times = self.times[position]
                # Compute delta t for a specific config
                configs_times[run_id] = (
                    elapsed,
                    True,
                )  # (time.time() - configs_times[run_id][0], True)
                # Update progress bar
                self.pbars[position].update()
                self.refresh()
            else:
                raise Exception(
                    f"Cannot parse type of message {type}, fix this."
                )

        except Exception as e:
            print(e)
            return

    def __enter__(self):
        """
        Needed when Progress Manager is used as context manager.
        Does nothing besides returning self.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Needed when Progress Manager is used as context manager.
        Closes the progress bar.
        """
        for pbar in self.pbars:
            pbar.close()


"""
Various options for random search model selection
"""


def choice(*args):
    """
    Implements a random choice between a list of values
    """
    return random.choice(args)


def uniform(*args):
    """
    Implements a uniform sampling given an interval
    """
    return random.uniform(*args)


def normal(*args):
    """
    Implements a univariate normal sampling given its parameters
    """
    return random.normalvariate(*args)


def randint(*args):
    """
    Implements a random integer sampling in an interval
    """
    return random.randint(*args)


def loguniform(*args):
    r"""
    Performs a log-uniform random selection.

    Args:
        *args: a tuple of (log min, log max, [base]) to use. Base 10 is used
            if the third argument is not available.

    Returns:
        a randomly chosen value
    """
    log_min, log_max, *base = args
    base = base[0] if len(base) > 0 else 10

    log_min = math.log(log_min) / math.log(base)
    log_max = math.log(log_max) / math.log(base)

    return base ** (random.uniform(log_min, log_max))



import pytest
import yaml

from pydgn.evaluation.grid import Grid
from pydgn.evaluation.random_search import RandomSearch


@pytest.fixture
def search_method_config_length():
    return [
        (Grid, "tests/evaluation/grid_search.yml", 6),
        (RandomSearch, "tests/evaluation/random_search.yml", 10),
    ]


def test_search_method(search_method_config_length):
    for search_method, filepath, num_of_configs in search_method_config_length:
        search = search_method(
            yaml.load(open(filepath, "r"), Loader=yaml.FullLoader)
        )
        # Check the amount of configurations expected and those produced
        # are the same
        assert len(search) == num_of_configs

        # No two configurations should be equal
        # (unless it's intended from the config file)
        for i in range(len(search)):
            for j in range(i + 1, len(search)):
                assert search[i] != search[j]



from torch.nn import Linear

from pydgn.model.interface import ModelInterface
from pydgn.static import LOSSES, SCORES, BEST_EPOCH, MIN, MAX
from pydgn.training.callback.early_stopping import PatienceEarlyStopper
from pydgn.training.event.state import State


class FakeModel(ModelInterface):
    def __init__(self):
        super().__init__(0, 0, 0, None, None)
        self.lin = Linear(10, 10)


def test_early_stopping_patience():

    for use_as_loss in [False, True]:
        for patience in [2, 10]:

            early_stopper = PatienceEarlyStopper(
                "validation_main_loss"
                if use_as_loss
                else "validation_main_score",
                mode=MIN if use_as_loss else MAX,
                patience=patience,
                checkpoint=False,
            )

            state = State(model=FakeModel(), optimizer=None, device="cpu")

            # Update state with epoch results
            epoch_results = {LOSSES: {}, SCORES: {}}
            state.update(epoch_results=epoch_results)

            num_epochs = 30
            for epoch in range(1, num_epochs + 1):
                state.update(epoch=epoch)

                state.epoch_results[LOSSES].update(
                    {f"validation_main_loss": epoch}
                )
                state.epoch_results[SCORES].update(
                    {f"validation_main_score": epoch}
                )

                early_stopper.on_epoch_end(state)

                if state.stop_training:
                    break

            if use_as_loss:
                # implies MIN is used. For this test we should stop and exit
                # the loop after patience epochs (the best epoch will always
                # be the first)
                assert state.best_epoch_results[BEST_EPOCH] == epoch - patience
            else:
                assert state.best_epoch_results[BEST_EPOCH] == num_epochs



from typing import Optional, List, Tuple

import torch
from torch_geometric.utils import to_dense_adj, to_dense_batch

from pydgn.model.interface import ReadoutInterface


class DotProductLinkReadout(ReadoutInterface):
    """
    Class that implements a simple readout mapping for
    link prediction via dot product
    """

    def forward(
        self, node_embeddings: torch.tensor, batch: torch.Tensor, **kwargs
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements a dot product scorer for link prediction

        Args:
            node_embeddings (`torch.Tensor`): the node embeddings
                of size `Nxd`
            batch (`torch.Tensor`): a tensor specifying to which graphs
                nodes belong to in the batch
            kwargs (dict): additional parameters (unused)

        Returns:
            a tuple (None, node_embeddings,
            [link scores, dense adjacency matrix])
        """
        edge_index = kwargs["edge_index"]

        z, _ = to_dense_batch(node_embeddings, batch)
        adj = to_dense_adj(edge_index, batch)

        batch_size, num_nodes, _ = z.size()
        """
        >>> As = torch.randn(3,2,5)
        >>> Bs = torch.randn(3,5,4)
        >>> torch.einsum('bij,bjk->bik', As, Bs) # batch matrix multiplication
        s = F.sigmoid(torch.einsum('bij,bji->bii', x, torch.transpose(x)))
        # batch matrix multiplication
        """
        return (
            None,
            node_embeddings,
            [torch.sigmoid(torch.matmul(z, z.transpose(1, 2))), adj],
        )



import torch
from torch.utils.data import sampler

from pydgn.data.dataset import DatasetInterface


class RandomSampler(sampler.RandomSampler):
    """
    This sampler wraps the dataset and saves the random permutation applied
    to the samples, so that it will be available
    for further use (e.g. for saving graph embeddings in the original order).
    The permutation is saved in the 'permutation' attribute.

    Args:
        data_source (:class:`pydgn.data.DatasetInterface`): the dataset object
    """

    def __init__(self, data_source: DatasetInterface):
        super().__init__(data_source)
        self.permutation = None

    def __iter__(self):
        """
        Iterates over the samples according to a pre-determined permutation.

        Returns:
             An iterable with permuted indices.

        """
        n = len(self.data_source)
        self.permutation = torch.randperm(n).tolist()
        return iter(self.permutation)



import argparse
import os
import sys

import gpustat
import yaml

from pydgn.static import *


def set_gpus(num_gpus):
    """
    Sets the visible GPUS for the experiments according to the availability
    in terms of memory. Prioritize GPUs with less memory usage.
    Sets the ``CUDA_DEVICE_ORDER`` env variable to ``PCI_BUS_ID``
    and ``CUDA_VISIBLE_DEVICES`` to the ordered list of GPU indices.

    Args:
        num_gpus: maximum number of GPUs to use when
            launching experiments in parallel
    """
    try:
        selected = []

        stats = gpustat.GPUStatCollection.new_query()

        for i in range(num_gpus):

            ids_mem = [
                res
                for res in map(
                    lambda gpu: (
                        int(gpu.entry["index"]),
                        float(gpu.entry["memory.used"])
                        / float(gpu.entry["memory.total"]),
                    ),
                    stats,
                )
                if str(res[0]) not in selected
            ]

            if len(ids_mem) == 0:
                # No more gpus available
                break

            best = min(ids_mem, key=lambda x: x[1])
            bestGPU, bestMem = best[0], best[1]
            # print(f"{i}-th best is {bestGPU} with mem {bestMem}")
            selected.append(str(bestGPU))

        print("Setting GPUs to: {}".format(",".join(selected)))
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(selected)
    except BaseException as e:
        print("GPU not available: " + str(e))


def evaluation(options: argparse.Namespace):
    """
    Takes the CLI arguments and launches the evaluation procedure.
    The method takes care of instantiating the Ray process and setting up
    the amount of resources available. Then it instantiates a
    :obj:`RiskAssesser` object to carry out the experiments.

    Args:
        :param options (argparse.Namespace):
    """
    kwargs = vars(options)
    debug = kwargs[DEBUG]
    configs_dict = yaml.load(
        open(kwargs[CONFIG_FILE], "r"), Loader=yaml.FullLoader
    )

    # Telegram bot settings
    telegram_config_file = configs_dict.get(TELEGRAM_CONFIG_FILE, None)
    if telegram_config_file is not None:
        telegram_config = yaml.load(
            open(telegram_config_file, "r"), Loader=yaml.FullLoader
        )
    else:
        telegram_config = None

    # Hardware initialization
    max_cpus = configs_dict[MAX_CPUS]

    device = configs_dict[DEVICE]
    use_cuda = CUDA in device

    if not use_cuda:
        max_gpus = 0
        gpus_per_task = 0
    else:
        # We will choose the GPU with least ratio of memory usage
        max_gpus = configs_dict[MAX_GPUS]
        # Choose which GPUs to use
        set_gpus(max_gpus)
        gpus_per_task = configs_dict[GPUS_PER_TASK]

    # we probably don't need this anymore, but keep it commented in case
    # OMP_NUM_THREADS = 'OMP_NUM_THREADS'
    # os.environ[OMP_NUM_THREADS] = "1"  # This is CRUCIAL to avoid
    # bottlenecks when running experiments in parallel. DO NOT REMOVE IT

    #
    # Once CUDA_VISIBLE_DEVICES has been set, we can start importing
    # all the necessary modules
    #
    import ray

    # we probably don't need this anymore, but keep it commented in case
    # Needed to avoid thread spawning, conflicts with multi-processing.
    # You may set a number > 1 but take into account the
    # number of processes on the machine
    # torch.set_num_threads(1)

    from pydgn.experiment.util import s2c
    from pydgn.data.splitter import Splitter
    from pydgn.evaluation.grid import Grid
    from pydgn.evaluation.random_search import RandomSearch

    assert GRID_SEARCH in configs_dict or RANDOM_SEARCH in configs_dict
    search_class = Grid if GRID_SEARCH in configs_dict else RandomSearch
    search = search_class(configs_dict)

    if use_cuda:
        # Ensure a generic "cuda" device is set when using more than 1 GPU
        search.device = CUDA

    # Set the bot token, chat_id configuration
    search.telegram_config = telegram_config

    # Set the random seed
    seed = search.seed if search.seed is not None else 42
    print(f"Base seed set to {seed}.")
    experiment = search.experiment
    experiment_class = s2c(experiment)
    exp_path = os.path.join(configs_dict[RESULT_FOLDER], f"{search.exp_name}")

    os.environ[PYDGN_RAY_NUM_GPUS_PER_TASK] = str(float(gpus_per_task))

    # You can make PyDGN work on a cluster of machines!
    if os.environ.get("ip_head") is not None:
        assert os.environ.get("redis_password") is not None
        ray.init(
            address=os.environ.get("ip_head"),
            _redis_password=os.environ.get("redis_password"),
        )
        print("Connected to Ray cluster.")
        print(f"Available nodes: {ray.nodes()}")
    # Or you can work on your single server
    else:
        ray.init(num_cpus=max_cpus, num_gpus=max_gpus)
        print(f"Started local ray instance.")

    data_splits_file = configs_dict[DATA_SPLITS_FILE]

    splitter = Splitter.load(data_splits_file)
    inner_folds, outer_folds = splitter.n_inner_folds, splitter.n_outer_folds
    print(
        f"Data splits loaded, outer folds are {outer_folds} "
        f"and inner folds are {inner_folds}"
    )

    # WARNING: leave the import here, it reads env variables set before
    from pydgn.evaluation.evaluator import RiskAssesser

    risk_assesser = RiskAssesser(
        outer_folds,
        inner_folds,
        experiment_class,
        exp_path,
        data_splits_file,
        search,
        final_training_runs=configs_dict[FINAL_TRAINING_RUNS],
        higher_is_better=search.higher_results_are_better,
        gpus_per_task=gpus_per_task,
        base_seed=seed,
    )

    risk_assesser.risk_assessment(debug=debug)
    ray.shutdown()


def get_args() -> argparse.Namespace:
    """
    Processes CLI arguments (i.e., the config file location and debug option)
    and returns a namespace.

    Returns:
        a namespace with the arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(CONFIG_FILE_CLI_ARGUMENT, dest=CONFIG_FILE)
    parser.add_argument(
        DEBUG_CLI_ARGUMENT, action="store_true", dest=DEBUG, default=False
    )
    return parser.parse_args()


def main():
    """
    Launches the experiment.
    """
    # Necessary to locate dotted paths in projects that use PyDGN
    sys.path.append(os.getcwd())

    options = get_args()
    try:
        evaluation(options)
    except Exception as e:
        raise e



import setuptools

__version__ = "1.2.5"

URL = "https://pydgn.readthedocs.io/en/latest/"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydgn",
    version=__version__,
    author="Federico Errica",
    author_email="f.errica@protonmail.com",
    description="A Python Package for Deep Graph Networks",
    long_description_content_type="text/markdown",
    url=URL,
    keywords=[
        "deep-graph-networks",
        "evaluation-framework",
        "deep-learning-for-graphs",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.8,<3.10",
    install_requires=[
        "PyYAML >= 5.4",
        "networkx >= 2.5",
        "requests >= 2.22.0",
        "matplotlib >= 3.3.4",
        "seaborn >= 0.9.0",
        "tensorboard >= 2.1.1",
        "tqdm >= 4.47.0",
        "ogb >= 1.3.3",
        "aioredis >= 1.3.1",
        "protobuf==3.20.2",
        "aiohttp == 3.7",
        "ray == 1.5.2",
        "gpustat >= 0.6.0",
        "torch >= 1.10.2",
        "torch-geometric >= 2.0.3",
        "torch-geometric-temporal >= 0.52.0",
        "wandb >= 0.12.15",
    ],
    packages=setuptools.find_packages(),
)



import os
import shutil
from pathlib import Path
from random import shuffle
from typing import List, Union, Tuple, Optional, Callable

import torch
import torch_geometric
from ogb.graphproppred import PygGraphPropPredDataset
from ogb.utils.url import decide_download, download_url, extract_zip
from torch_geometric.data import InMemoryDataset, Data, Batch
from torch_geometric.datasets import TUDataset, Planetoid
from torch_geometric_temporal import ChickenpoxDatasetLoader


class ZipDataset(torch.utils.data.Dataset):
    r"""
    This Dataset takes `n` datasets and "zips" them. When asked for the `i`-th
    element, it returns the `i`-th element of all `n` datasets.

    Args:
        datasets (List[torch.utils.data.Dataset]): An iterable with
            PyTorch Datasets

    Precondition:
        The length of all datasets must be the same

    """

    def __init__(self, datasets: List[torch.utils.data.Dataset], **kwargs):
        self.datasets = datasets
        assert len(set(len(d) for d in self.datasets)) == 1

    def __getitem__(self, index: int) -> List[object]:
        r"""
        Returns the `i`-th element of all datasets in a list

        Args:
            index (int): the index of the data point to retrieve from each
                dataset

        Returns:
            A list containing one element for each dataset in the ZipDataset
        """
        return [d[index] for d in self.datasets]

    def __len__(self) -> int:
        r"""
        Returns the length of the datasets (all of them have the same length)

        Returns:
            The unique number of samples of all datasets
        """
        return len(self.datasets[0])


class ConcatFromListDataset(InMemoryDataset):
    r"""
    Create a dataset from a list of :class:`torch_geometric.data.Data` objects.
    Inherits from :class:`torch_geometric.data.InMemoryDataset`

    Args:
        data_list (list): List of graphs.
    """

    def __init__(self, data_list: List[Data], **kwargs):
        super(ConcatFromListDataset, self).__init__("")
        self.data, self.slices = self.collate(data_list)

    def download(self):
        r"""
        Does nothing, the data list is already provided
        """
        pass

    def process(self):
        r"""
        Does nothing, the data list is already provided
        """
        pass

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        Does nothing, the data list is already provided
        """
        return []

    @property
    def raw_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        Does nothing, the data list is already provided
        """
        return []


class DatasetInterface(torch_geometric.data.dataset.Dataset):
    r"""
    Class that defines a number of properties essential to all datasets
    implementations inside PyDGN. These properties are used by the training
    engine and forwarded to the model to be trained. For some datasets, e.g.,
    :class:`torch_geometric.datasets.TUDataset`, implementing this interface
    is straightforward.

    Args:
        root (str): root folder where to store the dataset
        name (str): name of the dataset
        transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at run time
        pre_transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at dataset creation time
        pre_filter (Optional[Callable]): sample filtering to apply to each
            ``Data`` object at dataset creation time
    """

    def __init__(
        self,
        root: str,
        name: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        **kwargs,
    ):
        self.root = root
        self.name = name
        super().__init__(root, transform, pre_transform, pre_filter)

    @property
    def raw_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The name of the files in the :obj:`self.raw_dir` folder that must
        be present in order to skip downloading.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The name of the files in the :obj:`self.processed_dir` folder that
        must be present in order to skip processing.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    def download(self):
        r"""
        Downloads the dataset to the :obj:`self.raw_dir` folder.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    def process(self):
        r"""
        Processes the dataset to the :obj:`self.processed_dir` folder.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    def get(self, idx: int) -> Data:
        r"""
        Gets the data object at index :obj:`idx`.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    def len(self) -> int:
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )

    def __len__(self) -> int:
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        raise NotImplementedError(
            "You should subclass DatasetInterface and implement this method"
        )


class TemporalDatasetInterface(DatasetInterface):
    """
    Extension of DatasetInterface to the temporal scenario.
    """

    def get_mask(self, data: Union[Batch, Data]) -> torch.Tensor:
        """
        Computes the mask of time steps for which we need to make a prediction.

        Args:
            data: the data object

        Returns:
            A tensor indicating the time-steps at which we expect predictions
        """
        raise NotImplementedError(
            "You should subclass TemporalDatasetInterface and implement this "
            "method"
        )

    def get(self, idx: int) -> Data:
        """
        Gets element ``idx`` from object self.dataset

        Args:
            idx (int): the sample index

        Returns:

        """
        data = self.dataset[idx]
        data.time_prediction_mask = self.get_mask(data)
        return data


class IterableDatasetInterface(torch.utils.data.IterableDataset):
    r"""
    Class that implements the Iterable-style dataset, including multi-process
    data loading
    (https://pytorch.org/docs/stable/data.html#iterable-style-datasets).
    Useful when the dataset is too big and split in chunks of files to be
    stored on disk. Each chunk can hold a single sample or a set of samples,
    and there is the chance to shuffle sample-wise or chunk-wise.
    To get a subset of this dataset, just provide an argument `url_indices`
    specifying which chunks you want to use. Must be combined with an
    appropriate :class:`pydgn.data.provider.IterableDataProvider`.

    NOTE 1: We assume the splitter will split the dataset with respect to to
    the number of files stored on disk, so be sure that the length of your
    dataset reflects that number. Then, examples will be provided sequentially,
    so if each file holds more than one sample, we will still be able to create
    a batch of samples from one or multiple files.

    NOTE 2: NEVER override the __len__() method, as it varies dynamically with
    the ``url_indices`` argument.

    Args:
        root (str): root folder where to store the dataset
        name (str): name of the dataset
        transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at run time
        pre_transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at dataset creation time
        url_indices (Optional[List]): list of indices used to extract
            a portion of the dataset
    """

    def __init__(
        self,
        root: str,
        name: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        url_indices: Optional[List] = None,
        **kwargs,
    ):
        super().__init__()
        self.root = root
        self.name = name

        # This is needed to compute the a subset of the entire dataset
        if url_indices is None:
            processed_file_names = self.processed_file_names
        else:
            processed_file_names = [
                self.processed_file_names[i] for i in url_indices
            ]

        # This is needed to compute the length of a subset of the entire
        # dataset
        self.urls = processed_file_names
        self.shuffled_urls = self.urls

        self.start_index = 0
        self.end_index = len(self.shuffled_urls)

        # This information allows us to shuffle between urls and sub-patches
        # inside each url
        self._shuffle_urls = False
        self._shuffle_subpatches = False

        self.pre_transform = pre_transform
        self.transform = transform

        for p in self.raw_file_names:
            if not os.path.exists(os.path.join(self.raw_dir, p)):
                self.download()
                break

        if not os.path.exists(self.processed_dir):
            os.makedirs(os.path.join(self.processed_dir))

        for p in processed_file_names:
            if not os.path.exists(os.path.join(self.processed_dir, p)):
                self.process()
                break

    def shuffle_urls_elements(self, value: bool):
        r"""
        Shuffles elements contained in each file (associated with an url).
        Use this method when a single file stores multiple samples and you want
        to provide them in shuffled order.
        IMPORTANT: in this case we assume that each file contains a
        list of Data objects!

        Args:
            value (bool): whether or not to shuffle urls
        """
        self._shuffle_subpatches = value

    def shuffle_urls(self, value: bool):
        r"""
        Shuffles urls associated to individual files stored on disk

        Args:
            value (bool): whether or not to shuffle urls
        """
        self._shuffle_urls = value

        # Needed for multiple dataloader workers
        if self._shuffle_urls:
            shuffle(self.shuffled_urls)

    def splice(self, start: int, end: int):
        r"""
        Use this method to assign portions of the dataset to load to different
        workers, otherwise they will load the same samples.

        Args:
            start (int): the index where to start
            end (int): the index where to stop
        """
        self.start_index = start
        self.end_index = end

    def __iter__(self):
        r"""
        Generator that returns individual Data objects. If each files contains
        a list of data objects, these can be shuffled using the
        method :func:`shuffle_urls_elements`.

        Returns:
            a Data object with the next element to process
        """
        end_index = (
            self.end_index
            if self.end_index <= len(self.shuffled_urls)
            else len(self.shuffled_urls)
        )
        for url in self.shuffled_urls[self.start_index : end_index]:
            sample = torch.load(os.path.join(self.processed_dir, url))
            if isinstance(sample, list):
                if self._shuffle_subpatches:
                    shuffle(sample)

                data = [
                    self.transform(d) if self.transform is not None else d
                    for d in sample
                ]
            else:
                data = [
                    self.transform(sample)
                    if self.transform is not None
                    else sample
                ]

            for i in range(len(data)):
                yield data[i]

    @property
    def raw_file_names(self) -> List[Union[str, Path]]:
        r"""
        The name of the files in the :obj:`self.raw_dir` folder that must
        be present in order to skip downloading.
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    @property
    def raw_dir(self) -> Union[str, Path]:
        r"""
        The path where the raw data should be downloaded
        :return: a string
        """
        return os.path.join(self.root, self.name, "raw")

    @property
    def raw_paths(self) -> List[Union[str, Path]]:
        r"""
        The absolute filepaths that must be present in order to skip
        downloading.
        """
        files = self.raw_file_names
        return [os.path.join(self.raw_dir, f) for f in files]

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The list of file names that must be present in order to skip
        downloading.
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    @property
    def processed_dir(self) -> Union[str, Path]:
        r"""
        The folder where to store processed data files.
        """
        return os.path.join(self.root, self.name, "processed")

    @property
    def processed_paths(self) -> List[Union[str, Path]]:
        r"""
        The absolute filepaths that must be present in order to skip
        processing.
        """
        files = self.processed_file_names
        return [os.path.join(self.processed_dir, f) for f in files]

    def download(self):
        r"""
        Downloads the dataset to the :obj:`self.raw_dir` folder."""
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    def process(self):
        r"""
        Processes the dataset to the :obj:`self.processed_dir` folder.
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    def get(self, idx: int) -> Data:
        r"""
        Gets the data object at index :obj:`idx`.
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing,
        but in the end it depends on the model that is implemented).
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing,
        but in the end it depends on the model that is implemented).
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        raise NotImplementedError(
            "You should subclass IterableDatasetInterface "
            "and implement this method"
        )

    def __len__(self):
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply
        with PyG interface
        """
        # It's important it stays dynamic, because
        # self.urls depends on url_indices
        return len(self.urls)


class TUDatasetInterface(TUDataset):
    r"""
    Class that wraps the :class:`torch_geometric.datasets.TUDataset` class to
    provide aliases of some fields. It implements the interface
    ``DatasetInterface`` but does not extend directly to avoid
    clashes of ``__init__`` methods
    """

    def __init__(
        self,
        root,
        name,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        **kwargs,
    ):
        self.name = name
        # Do not call DatasetInterface __init__ method in this case, because
        # otherwise it will break
        super().__init__(
            root=root,
            name=name,
            transform=transform,
            pre_transform=pre_transform,
            pre_filter=pre_filter,
        )

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.num_node_features

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.num_edge_features

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        return self.num_classes

    def download(self):
        r"""
        Downloads the TUDataset dataset to the :obj:`self.raw_dir` folder."""
        super().download()

    def process(self):
        r"""
        Processes the TUDataset dataset to the :obj:`self.processed_dir` folder
        """
        super().process()


class PlanetoidDatasetInterface(Planetoid):
    r"""
    Class that wraps the :class:`torch_geometric.datasets.Planetoid` class to
    provide aliases of some fields. It implements the interface
    ``DatasetInterface`` but does not extend directly to avoid clashes
    of ``__init__`` methods
    """

    def __init__(
        self,
        root,
        name,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        **kwargs,
    ):
        self.name = name
        # Do not call DatasetInterface __init__ method in this case, because
        # otherwise it will break
        super().__init__(
            root=root,
            name=name,
            transform=transform,
            pre_transform=pre_transform,
        )

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.num_node_features

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.num_edge_features

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        return self.num_classes

    def download(self):
        r"""
        Downloads the Planetoid dataset to the :obj:`self.raw_dir` folder."""
        super().download()

    def process(self):
        r"""
        Processes the Planetoid dataset to the :obj:`self.processed_dir` folder
        """
        super().process()

    def __len__(self) -> int:
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        if isinstance(self.data, Data):
            return 1
        return len(self.data)


class OGBGDatasetInterface(PygGraphPropPredDataset):
    r"""
    Class that wraps the :class:`ogb.graphproppred.PygGraphPropPredDataset`
    class to provide aliases of some fields. It implements the interface
    ``DatasetInterface`` but does not extend directly to avoid clashes
    of ``__init__`` methods
    """

    def __init__(
        self,
        root,
        name,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        meta_dict=None,
        **kwargs,
    ):
        super().__init__(
            "-".join(name.split("_")),
            root,
            transform,
            pre_transform,
            meta_dict,
        )
        self.name = name
        self.data.y = self.data.y.squeeze()

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.data.x.shape[1]

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        if self.data.edge_attr is not None:
            return self.data.edge_attr.shape[1]
        else:
            return 0

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        return self.data.y.shape[1]

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The list of file names that must be present in order to skip
        downloading.
        """
        return ["data.pt"]

    def download(self):
        r"""
        Downloads the OGB dataset to the :obj:`self.raw_dir` folder.
        """
        url = self.meta_info["url"]
        if decide_download(url):
            path = download_url(url, self.original_root)
            extract_zip(path, self.original_root)
            print(f"Removing {path}")
            os.unlink(path)
            print(f"Removing {self.root}")
            shutil.rmtree(self.root)
            print(
                f"Moving "
                f"{os.path.join(self.original_root, self.download_name)} "
                f"to {self.root}"
            )
            shutil.move(
                os.path.join(self.original_root, self.download_name), self.root
            )

    def process(self):
        r"""
        Processes the OGB dataset to the :obj:`self.processed_dir` folder.
        """
        super().process()

    def __len__(self) -> int:
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        return self.data.y.shape[0]


class ToyIterableDataset(IterableDatasetInterface):
    r"""
    Class that implements the Iterable-style dataset, including multi-process
    data loading
    (https://pytorch.org/docs/stable/data.html#iterable-style-datasets).
    Useful when the dataset is too big and split in chunks of files to be
    stored on disk. Each chunk can hold a single sample or a set of samples,
    and there is the chance to shuffle sample-wise or chunk-wise. To get a
    subset of this dataset, just provide an argument `url_indices` specifying
    which chunks you want to use. Must be combined with an appropriate
    :class:`pydgn.data.provider.IterableDataProvider`.

    Args:
        root (str): root folder where to store the dataset
        name (str): name of the dataset
        transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at run time
        pre_transform (Optional[Callable]): transformations to apply to each
            ``Data`` object at dataset creation time
        url_indices (Optional[List]): list of indices used to extract a
            portion of the dataset
    """

    def __init__(
        self,
        root: str,
        name: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        url_indices: Optional[List] = None,
        **kwargs,
    ):
        self.num_files = 5
        self.url_indices = [f"{i}" for i in range(self.num_files)]
        super().__init__(
            root, name, transform, pre_transform, url_indices, **kwargs
        )

    @property
    def raw_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The name of the files in the :obj:`self.raw_dir` folder that must
        be present in order to skip downloading.
        """
        return []

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        r"""
        The list of file names that must be present in order to skip
        downloading.
        """
        return [f"fake_processed_{i}.pt" for i in range(self.num_files)]

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return 5

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return 0

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        return 1

    def download(self):
        r"""
        Does nothing, the data list is already provided
        """
        pass

    def process(self):
        r"""
        Creates a fake dataset and stores it to the :obj:`self.processed_dir`
        folder. Each file will contain a list of 10 fake graphs.
        """
        for i in range(len(self)):
            fake_graphs = []
            for g in range(10):
                fake_graphs.append(
                    Data(
                        x=torch.zeros(20, 5),
                        y=torch.zeros(1, 1),
                        edge_index=torch.zeros(2, 1).long(),
                    )
                )
            torch.save(fake_graphs, self.processed_paths[i])


class ChickenpoxDatasetInterface(TemporalDatasetInterface):
    r"""
    Class that acts as wrapper to Pytorch Geometric Temporal Chickenpox dataset

    Args:
        root (Union[str,Path]): root data folder
        name (str): name of the dataset
        lags: (how many original timesteps to aggregate into
            a single time step)
        kwargs: optional additional parameters to `ChickenpoxDatasetLoader`
            (from PyG Temporal)
    """

    def __init__(self, root, name, lags=4, **kwargs):
        super().__init__(root, name, **kwargs)

        self.lags = lags
        self.dataset = ChickenpoxDatasetLoader().get_dataset(lags=lags)

    def get_mask(self, data):
        """
        Computes the mask of time steps for which we need to make a prediction.
        In this case data is a Data object containing a snapshot of a
        single graph sequence, and the task is node classification
        at each time step

        Args:
            data: the data object

        Returns:
            A tensor indicating the time-steps at which we expect predictions
        """
        mask = torch.ones((1, 1))  # time_steps x 1
        return mask

    @property
    def dim_node_features(self) -> int:
        r"""
        Specifies the number of node features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return self.dataset.features[0].shape[1]

    @property
    def dim_edge_features(self) -> int:
        r"""
        Specifies the number of edge features (after pre-processing, but in
        the end it depends on the model that is implemented).
        """
        return 0

    @property
    def dim_target(self) -> int:
        r"""
        Specifies the dimension of each target vector.
        """
        return 1

    def len(self):
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        return len(self)

    def __len__(self):
        r"""
        Returns the number of graphs stored in the dataset.
        Note: we need to implement both `len` and `__len__` to comply with
        PyG interface
        """
        return len(self.dataset.features)  # see DynamicGraphTemporalSignal
